let magic_tag = 0x51415030

(* Returns two lists of wires, separated by bank: shared (public) and local (private). *)
let banks qap =
  Maps.PTree.fold
  (fun (s, l) w b ->
    match b with
    | QAP.Bank_Shared _ -> (w :: s, l)
    | QAP.Bank_Local -> (s, w :: l)
  )
  (QAP.meta_data qap)
  ([], [])

let print_z oc z = Printf.fprintf oc "%s" Camlcoq.Z.(hex z)
let positive oc p = print_z oc Camlcoq.Z.(Zpos p)
let print_N oc n = print_z oc (Camlcoq.Z.of_N n)

let print_list oc l =
  Printf.fprintf oc "(";
  List.iter (Printf.fprintf oc "%a " positive) l;
  Printf.fprintf oc ")"

let sizes oc qap =
  let (s, l) = banks qap in
  Printf.fprintf oc "(%a %a)"
  print_list s
  print_list l

let tree_size p =
  Maps.PTree.fold
  (fun a _ _ -> a + 1)
  p
  0

let lc_size p : int =
  QAP.LC.fold
  (ModpProof.zpEquiv LargePrime.large_prime)
  (ModpProof.zpPlus LargePrime.large_prime)
  (ModpProof.zpZero LargePrime.large_prime)
  (fun a _ _ -> a + 1)
  p
  0

let int_of_bank =
  function
  | QAP.Bank_Shared _ -> 1
  | QAP.Bank_Local -> 2

let bank oc b =
  Printf.fprintf oc "%x" (int_of_bank b)

let poly oc poly : unit =
  Maps.PTree.fold
  (fun () r lc ->
    QAP.LC.fold
    (ModpProof.zpEquiv LargePrime.large_prime)
    (ModpProof.zpPlus LargePrime.large_prime)
    (ModpProof.zpZero LargePrime.large_prime)
    (fun () w k ->
      Printf.fprintf oc "(%a %a %a)\n"
      positive r
      positive w
      print_N k
    )
    lc
    ()
  )
  poly
  (Printf.fprintf oc "(");
  (Printf.fprintf oc ")")

let output_qap (oc: out_channel) qap : unit =
  Printf.fprintf oc
  "(%a %a %a %a %a)\n"
   positive (QAP.next_root qap)
   sizes qap
   poly (QAP.cV qap)
   poly (QAP.cW qap)
   poly (QAP.cY qap)

let input_lp ic : unit = Scanf.fscanf ic " ( " ()
let input_rp ic : unit = Scanf.fscanf ic " ) " ()

let input_eqn ic : (int * int * int) option =
  try
    Scanf.fscanf ic " (%x %x %x)" (fun r w k -> Some (r, w, k))
  with Scanf.Scan_failure _ -> None

let input_poly ic (cb: int * int * int -> unit) : unit =
  input_lp ic;
  let rec doit () =
    match input_eqn ic with
    | Some q -> q |> cb |> doit
    | None -> ()
  in doit ();
  input_rp ic

let input_qap (ic: in_channel)
  (nroots: int -> unit)
  (banks : int -> int -> int -> unit)
  polyV polyW polyY
  =
  input_lp ic;
  Scanf.fscanf ic "%x" nroots;
  Scanf.fscanf ic " (%x %x %x)" banks;
  input_poly ic polyV;
  input_poly ic polyW;
  input_poly ic polyY;
  input_rp ic

let int_split i =
  let b0 = (i lsr 0o00) land 0xff in
  let b1 = (i lsr 0o10) land 0xff in
  let b2 = (i lsr 0o20) land 0xff in
  let b3 = (i lsr 0o30) land 0xff in
  (b3, b2, b1, b0)

let int_merge b3 b2 b1 b0 : int =
  (b3 lsl 0o30) lor
  (b2 lsl 0o20) lor
  (b1 lsl 0o10) lor
  (b0 lsl 0o00)

let output_int32be oc (i: int) : unit =
  let o c = c |> char_of_int |> output_char oc in
  let (b3, b2, b1, b0) = int_split i in
  o b3 ; o b2 ; o b1 ; o b0

let input_int32be ic : int =
  let i () = input_char ic |> int_of_char in
  let b3 = i() in
  let b2 = i() in
  let b1 = i() in
  let b0 = i() in
  int_merge b3 b2 b1 b0

let output_Z oc z =
  z |> Camlcoq.Z.to_int |> output_int32be oc

let output_positive oc p =
  Camlcoq.Z.Zpos p |> output_Z oc

let input_positive ic =
  input_int32be ic |> Camlcoq.P.of_int

let input_Z ic =
  Camlcoq.Z.Zpos(input_positive ic)

let output_zp oc z =
  let (((((((b7, b6), b5), b4), b3), b2), b1), b0) = Project.split8 z in
  let o i = i |> Camlcoq.Z.of_N |> Camlcoq.Z.to_int |> output_int32be oc in
  o b7 ; o b6 ; o b5 ; o b4 ; o b3 ; o b2 ; o b1 ; o b0

let input_zp ic =
  let i () = input_int32be ic |> Camlcoq.N.of_int in
  let b7 = i () in
  let b6 = i () in
  let b5 = i () in
  let b4 = i () in
  let b3 = i () in
  let b2 = i () in
  let b1 = i () in
  let b0 = i () in
  Project.merge8 (((((((b7, b6), b5), b4), b3), b2), b1), b0)

let output_poly oc p =
  p |> tree_size |> output_int32be oc;
  Maps.PTree.fold
  (fun () r lc ->
    lc |> lc_size |> output_int32be oc;
    QAP.LC.fold
    (ModpProof.zpEquiv LargePrime.large_prime)
    (ModpProof.zpPlus LargePrime.large_prime)
    (ModpProof.zpZero LargePrime.large_prime)
    (fun () w k ->
      r |> output_positive oc;
      w |> output_positive oc;
      k |> output_zp oc
    )
    lc
    ()
  )
  p
  ()

let input_poly ic cb s =
  let state = ref s in
  let sz = input_int32be ic in
  for i = 1 to sz do
    let lc_sz = input_int32be ic in
    for lc = 1 to lc_sz do
      let r = input_positive ic in
      let w = input_positive ic in
      let k = input_zp ic in
      state := cb !state r w k
    done
  done;
  !state

let output_qap_bin oc qap : unit =
  output_int32be oc magic_tag; (* magic *)
  output_positive oc (QAP.next_root qap); (* roots *)
  output_int32be oc 2; (* number of banks *)
  let (s, l) = banks qap in
  List.iter
  (fun b ->
    b |> List.length |> output_int32be oc; (* bank size *)
    b |> List.iter (output_positive oc)    (* wires in the bank *)
  )
  [ s ; l ];
  qap |> QAP.cV |> output_poly oc; (* V *)
  qap |> QAP.cW |> output_poly oc; (* W *)
  qap |> QAP.cY |> output_poly oc; (* Y *)
  ()

let input_qap_bin ic cb_init cb_eqn =
  let magic = input_int32be ic in
  assert (magic = magic_tag);
  let nroot = input_positive ic in
  let nbanks = input_int32be ic in
  let bank_wires = Array.init nbanks (fun i ->
    let sz = input_int32be ic in
    Array.init sz (fun w -> input_int32be ic)
  ) in
  let state = ref (cb_init nroot bank_wires) in
  for p = 0 to 2 do
    state := input_poly ic (fun s -> cb_eqn s p) !state
  done;
  !state
