open Camlcoq

let rec omega = Datatypes.S omega
type run_mode = Int32 | QAP
let option_run : run_mode option ref = ref None
let option_input_public : string ref = ref ""
let option_input_private: string ref = ref ""
let option_tc : bool ref = ref true
let option_cse : bool ref = ref true
let option_cp : bool ref = ref true
let option_dc : bool ref = ref true

let string_of_run_mode =
  function
  | Int32 -> "eval"
  | QAP -> "qap"

let openInput (fn: string) =
  begin
    try
      let fd = open_in_bin fn in
      object
        method close = close_in fd
        method next_int =
          try
            input_line fd |> int_of_string
          with
          | End_of_file ->
             failwith ("no next int in " ^ fn)
      end
    with _ ->
      (* no file, just return random numbers *)
      object
        method close = ()
        method next_int = Random.int 10
      end
  end

(* Printing a behavior *)

let string_of_int i = i |> Integers.Int.unsigned |> Z.hex
let print_Z oc z = Printf.fprintf oc "%s" (Z.hex z)
let print_N oc n = print_Z oc (match n with N.N0 -> Z.Z0 | N.Npos p -> Z.Zpos p)

let string_of_constant =
  function
  | Expr.ConstInt i -> string_of_int i
  | Expr.ConstStackAddr ofs -> Printf.sprintf "stk[%s]" (string_of_int ofs)
  | Expr.ConstGlobal (s, ofs) -> Printf.sprintf "%s[%s]" (extern_atom s) (string_of_int ofs)

let print_val pv oc =
  function
  | AbStateImpl.Const c -> Printf.fprintf oc "%s" (string_of_constant c)
  | AbStateImpl.Var v -> Printf.fprintf oc "%a" pv v

let string_of_chunk =
  function
  | AST.Mint8signed -> "Mint8s"
  | AST.Mint8unsigned -> "Mint8u"
  | AST.Mint16signed -> "Mint16s"
  | AST.Mint16unsigned -> "Mint16u"
  | AST.Mint32 -> "Mint32"
  | AST.Mint64 -> "Mint64"
  | AST.Mfloat32 -> "Mfloat32"
  | AST.Mfloat64 -> "Mfloat64"
  | AST.Many32 -> "Many32"
  | AST.Many64 -> "Many64"

let string_of_list f ?sep:(s="; ") l =
  l |> List.map f |> String.concat s

let print_list (p: out_channel -> 'a -> unit) ?sep:(s="; ") (oc: out_channel) (l: 'a list): unit =
  l
  |> List.fold_left (fun b i -> Printf.fprintf oc "%s%a" (if b then s else "") p i; true) false
  |> (ignore : bool -> unit)

let print_event pv oc =
  function
  | AbState.AEinput (k, f, ofs, v) ->
      Printf.fprintf oc "vload %s %s[%s] → %a"
      (string_of_chunk k)
      (extern_atom f)
      (string_of_int ofs)
      (print_val pv) v
  | AbState.AEannot (msg, args) ->
      Printf.fprintf oc "annot %s(%a)"
      (camlstring_of_coqstring msg)
      (print_list (fun oc i -> Printf.fprintf oc "%s" (string_of_int i)) ~sep:", ") args
  | AbState.AEoutput (k, f, ofs, v) ->
      Printf.fprintf oc "vstore %s %s[%s] ← %a"
      (string_of_chunk k)
      (extern_atom f)
      (string_of_int ofs)
      (print_val pv) v

let print_trace oc pv tr : unit =
  tr |>
  List.iter (fun i -> Printf.fprintf oc "%a\n" (print_event pv) i)

let print_beh verbose pv p oc ((tr, r), s) =
  if verbose
  then (
    print_trace oc (pv (snd s)) tr;
    Printf.fprintf oc "Return: %a\n" (print_val (pv (snd s))) r
  );
  p s;
  ()

let string_of_Z_coeff z =
  if Z.compare Z.one z = 0
  then ""
  else Printf.sprintf "%s · " (Z.hex z)

let print_lc qap oc lc =
  let lc = match Maps.PTree.get lc (QAP.env qap) with
  | Some (Datatypes.Coq_inl (lc, rng)) -> lc
  | Some (Datatypes.Coq_inr br) -> failwith "print_lc: binary repr"
  | None -> assert false in
  if
    QAP.LC.fold
    (ModpProof.zpEquiv LargePrime.large_prime)
    (ModpProof.zpPlus LargePrime.large_prime)
    (ModpProof.zpZero LargePrime.large_prime)
  (fun first x k ->
    Printf.fprintf oc "%s%s[%d]" (if first then "" else " + ") (string_of_Z_coeff (Z.of_N k)) (P.to_int x);
    false
    )
  lc
  true
  then Printf.fprintf oc "0"

let print_qap verbose qap =
  if verbose
  then (
  let z x = if x = "" then "0" else x in
  QAP.fold_equations
    (ModpProof.zpEquiv LargePrime.large_prime)
    (ModpProof.zpPlus LargePrime.large_prime)
    (ModpProof.zpZero LargePrime.large_prime)
  (fun v w y -> Printf.sprintf "(%s) × (%s) = %s" (z v) (z w) (z y))
  (fun acc x k ->
    Printf.sprintf "%s%s%s[%d]" acc (if acc = "" then "" else " + ")
    (string_of_Z_coeff (Z.of_N k))
    (P.to_int x))
  ""
  qap
  |> List.rev
  |> List.iter print_endline
  );
  let oc = open_out_bin "qap.bin" in
  QapIO.output_qap_bin oc qap;
  close_out oc

(* Binary decomposition of a ZP on [n] bits. [n] should be large enough. *)
let bits (f: bool -> 'a) (n0: int) (z: Modp.zp) (next: 'a list) : 'a list =
  let rec bits n z acc =
  BinNums.(
  begin match n, z with
  | 0, N0 -> List.rev_append acc next
  | 0, _ -> failwith (Printf.sprintf "bits: underflow %d" n0)
  | _, N0 -> bits (n - 1) z (f false :: acc)
  | _, Npos Coq_xH -> bits (n - 1) N0 (f true :: acc)
  | _, Npos (Coq_xO z') -> bits (n - 1) (Npos z') (f false :: acc)
  | _, Npos (Coq_xI z') -> bits (n - 1) (Npos z') (f true :: acc)
  end)
  in
  bits n0 z []
