Require Import Monads Interval.
Import Integers.
Import MoreLib Interval.Notations.
Import Utf8 ZArith String.

Unset Elimination Schemes.
Set Implicit Arguments.

Local Open Scope interval_scope.

Inductive input_data := InputNow (compile_time: bool) (i: int) | InputLater.

(* Hints in Gepetto:
 type hint =
    | Split of bool * plan // Shall I split? What is the representation?
    | Mlt1 | Mlt0          // Shall I record the result of this multiplication? Cheaper than Mlt of bool
    | Zerop of bool        // Shall I zerop (using 2 variables)?
    | Truncate of bool     // Shall I truncate?
    | Output of bool       // Shall I assign?
    | Nroot of int         // How many equations so far?
    | IsBit of bool        // Is this int always in 0..1?
    | Empty
*)

Module Arith.
Section S.
Variables S E c : Type.
Variable incr : S → S → Prop.
Record t :=
{ of_int : int → STm.t incr E c
; to_int: c → STm.t incr E int
; add : c → c → STm.t incr E c
; sub : c → c → STm.t incr E c
(* Multiplication: the boolean is a hint for the prover that a fresh wire is to be introduced. *)
; mul : c → c → STm.t incr E (c * bool)
(* Logical/bitwise operations; hints say, for each argument:
   - if we need a binary decomposition or not
   - if we need a binary decomposition, the number of bits *)
; land : c → c → STm.t incr E (c * (option nat * option nat))
; lor : c → c → STm.t incr E (c * (option nat * option nat))
; lxor : c → c → STm.t incr E (c * (option nat * option nat))
(* ; lnot : c → STm.t incr E (c * (option nat)) *)
(* ; shl : c → int → STm.t incr E c *)
; shru : c → int → STm.t incr E (c * (option nat))
; cmp : bool → comparison → c → c → STm.t incr E c
; init : S
; input : input_data → option (int * int) → STm.t incr E c
; output : c → STm.t incr E (c * option nat)
(* Optionally returns an inverse after shifting by n×Int.modulus. *)
; assert : c → STm.t incr E (option (c * nat))
; assert_eq : c → c → STm.t incr E unit
}.

Variable a : t.
Variable solution : S → (c → Z) → Prop.
Variable Inv : S → Prop.

(* solution should be compatible with incr for this spec to make sense. *)
Record spec : Prop :=
{
of_int_sound : ∀ {i σ x σ' Hincr},
  of_int a i σ = STm.Result x σ' Hincr →
  ∀ {ρ}, solution σ' ρ →
    Inv σ →
    Int.repr (ρ x) = i (* maybe too strong *)
    ∧ Inv σ'
;
to_int_sound : ∀ {x σ i σ' Hincr},
  to_int a x σ = STm.Result i σ' Hincr →
  ∀ {ρ}, solution σ' ρ →
    Inv σ →
    Int.repr (ρ x) = i
    ∧ Inv σ'
;
add_sound : ∀ {x y σ z σ' Hincr},
  add a x y σ = STm.Result z σ' Hincr →
  ∀ {ρ},
    solution σ' ρ →
    Inv σ →
    Int.add (Int.repr (ρ x)) (Int.repr (ρ y)) = (Int.repr (ρ z))
    ∧ Inv σ'
;
sub_sound : ∀ {x y σ z σ' Hincr},
  sub a x y σ = STm.Result z σ' Hincr →
  ∀ {ρ},
    solution σ' ρ →
    Inv σ →
    Int.sub (Int.repr (ρ x)) (Int.repr (ρ y)) = (Int.repr (ρ z))
    ∧ Inv σ'
;
mul_sound : ∀ {x y σ z b σ' Hincr},
  mul a x y σ = STm.Result (z, b) σ' Hincr →
  ∀ {ρ},
    solution σ' ρ →
    Inv σ →
    Int.mul (Int.repr (ρ x)) (Int.repr (ρ y)) = (Int.repr (ρ z))
    ∧ Inv σ'
;
land_sound : ∀ {x y σ z h σ' Hincr},
  land a x y σ = STm.Result (z, h) σ' Hincr →
  ∀ {ρ},
    solution σ' ρ →
    Inv σ →
    Int.and (Int.repr (ρ x)) (Int.repr (ρ y)) = (Int.repr (ρ z))
    ∧ Inv σ'
;
lor_sound : ∀ {x y σ z h σ' Hincr},
  lor a x y σ = STm.Result (z, h) σ' Hincr →
  ∀ {ρ},
    solution σ' ρ →
    Inv σ →
    Int.or (Int.repr (ρ x)) (Int.repr (ρ y)) = (Int.repr (ρ z))
    ∧ Inv σ'
;
lxor_sound : ∀ {x y σ z h σ' Hincr},
  lxor a x y σ = STm.Result (z, h) σ' Hincr →
  ∀ {ρ},
    solution σ' ρ →
    Inv σ →
    Int.xor (Int.repr (ρ x)) (Int.repr (ρ y)) = (Int.repr (ρ z))
    ∧ Inv σ'
;
(*
lnot_sound : ∀ {x σ z h σ' Hincr},
  lnot a x σ = STm.Result (z, h) σ' Hincr →
  ∀ {ρ},
    solution σ' ρ →
    Inv σ →
    Int.not (Int.repr (ρ x)) = (Int.repr (ρ z))
    ∧ Inv σ'
;
*)
(*
shl_sound : ∀ {x i σ z σ' Hincr},
  shl a x i σ = STm.Result z σ' Hincr →
  ∀ {ρ},
    solution σ' ρ →
    Inv σ →
    Int.shl (Int.repr (ρ x)) i = (Int.repr (ρ z))
    ∧ Inv σ'
;
*)
shru_sound : ∀ {x i σ z h σ' Hincr},
  shru a x i σ = STm.Result (z, h) σ' Hincr →
  ∀ {ρ},
    solution σ' ρ →
    Inv σ →
    Int.shru (Int.repr (ρ x)) i = (Int.repr (ρ z))
    ∧ Inv σ'
;
cmp_sound : ∀ {s c x y σ z σ' Hincr},
  cmp a s c x y σ = STm.Result z σ' Hincr →
  ∀ {ρ},
    solution σ' ρ →
    Inv σ →
    (if (if s then Int.cmp else Int.cmpu) c (Int.repr (ρ x)) (Int.repr (ρ y)) then Int.one else Int.zero) = (Int.repr (ρ z))
    ∧ Inv σ'
;
init_sound : Inv (init a)
;
input_sound : ∀ {x r σ y σ' Hincr},
  input a x r σ = STm.Result y σ' Hincr →
  Inv σ →
  Inv σ'
;
output_sound : ∀ {x σ y h σ' Hincr},
  output a x σ = STm.Result (y, h) σ' Hincr →
  ∀ {ρ},
    solution σ' ρ →
    Inv σ →
    Int.repr (ρ y) = Int.repr (ρ x)
    ∧ Inv σ'
;
assert_sound : ∀ {x y σ σ' Hincr},
  assert a x σ = STm.Result y σ' Hincr →
  ∀ {ρ},
    solution σ' ρ →
    Inv σ →
    Int.eq (Int.repr (ρ x)) Int.zero = false
    ∧ Inv σ'
;

assert_eq_sound : ∀ {x y σ σ' Hincr},
  assert_eq a x y σ = STm.Result tt σ' Hincr →
  ∀ {ρ},
    solution σ' ρ →
    Inv σ →
    Int.eq (Int.repr (ρ x)) (Int.repr (ρ y)) = true
    ∧ Inv σ'
}.

Import STm.Notations.
Context {incr_refl: RelationClasses.Reflexive incr}
        {incr_trans: RelationClasses.Transitive incr}.

(** Derived operation: decide if a signed integer is negative.
      (x < 0) = (x & 2³¹ >> 31)
*)
Definition is_neg (x: c) : STm.t incr E (c * ((option nat * c) * (option nat * c) * (option nat * c))) :=
  do mask ← of_int a (Int.repr Int.half_modulus);
  do ih ← land a x mask;
  let '(i, (h,h'')) := ih in
  do rh' ← shru a i (Int.repr 31);
  let '(r, h') := rh' in
  STm.ret (r, ((h, x), (h'', mask), (h', i))).

Context (a_sound: spec).
Context (incr_solution : ∀ σ σ', incr σ σ' → solution σ' ⊆ solution σ).
Hint Resolve incr_solution.

Definition is_neg_sound {x y h σ σ' Hincr} :
  is_neg x σ = STm.Result (y, h) σ' Hincr →
  ∀ {ρ},
    solution σ' ρ →
    Inv σ →
    (if Int.lt (Int.repr (ρ x)) Int.zero then Int.one else Int.zero) = Int.repr (ρ y)
    ∧ Inv σ'.
Proof.
  unfold is_neg.
  intros H ρ H0 H1.
  hsplit. subst.
  repeat match goal with
  | H : of_int _ _ _ = STm.Result _ ?s _, K : solution ?s' ?ρ |- _ =>
    let SOL := fresh in
    assert (solution s ρ) as SOL by eauto;
    let i := fresh in
    destruct (of_int_sound a_sound H SOL); clear H; eauto
  | H : land _ _ _ _ = STm.Result _ ?s _, K : solution ?s' ?ρ |- _ =>
    let SOL := fresh in
    assert (solution s ρ) as SOL by eauto;
    destruct (land_sound a_sound H SOL); clear H; eauto
  | H : shru _ _ _ _ = STm.Result _ ?s _, K : solution ?s' ?ρ |- _ =>
    let SOL := fresh in
    assert (solution s ρ) as SOL by eauto;
    destruct (shru_sound a_sound H SOL); clear H; eauto
  | H : Int.repr (ρ ?x) = _ |- _ => rewrite H in *; clear H
  | H : _ = Int.repr (ρ ?x) |- _ => rewrite <- H in *; clear H
  end.
  split; eauto.
  generalize (Int.repr (ρ h4)). clear.
  intros i.
  rewrite <- Int.and_shru.
  change (Int.shru (Int.repr Int.half_modulus) (Int.repr 31)) with Int.one.
  unfold Int.lt. change (Int.signed Int.zero) with Z0.
  destruct (Coqlib.zlt _ _) as [ NEG | POS ];
  apply Int.same_bits_eq; intros n Hn;
  rewrite (Int.bits_and _ _ _ Hn);
  rewrite (Int.bits_shru _ _ _ Hn);
  clear Hn;
  change (Int.unsigned (Int.repr 31)) with 31%Z;
  rewrite Int.bits_one;
  (destruct (Coqlib.zeq _ _) as [ -> | Hnz ];
    [ rewrite Bool.andb_true_r | rewrite Bool.andb_false_r; try rewrite Int.bits_zero; reflexivity ]).
  unfold Int.signed in *.
  destruct (Coqlib.zlt (Int.unsigned _) _). generalize (Int.unsigned_range i); Psatz.lia.
  unfold Int.testbit. rewrite Z.testbit_odd. rewrite test_sign_bit. reflexivity. Psatz.lia.
  rewrite Int.bits_zero.
  unfold Int.signed in *.
  destruct (Coqlib.zlt (Int.unsigned _) _). 2: generalize (Int.unsigned_range i); Psatz.lia.
  unfold Int.testbit. rewrite (Zbits.Ztestbit_above 31). reflexivity. change (two_power_nat 31) with Int.half_modulus. Psatz.lia.
  Psatz.lia.
Qed.

End S.
End Arith.
