Require Import Utf8.
Require Import Expr Monads.
Import Relation_Definitions.
Import Coqlib Events Registers Integers Globalenvs Memory.

Unset Elimination Schemes.
Set Implicit Arguments.

(* This state abstraction uses variables internally, not to be confused with program variables, aka registers. *)

Section STATE.

Context (ab_state: Type) (incr: relation ab_state).

Record state_abstraction (E var: Type) :=
  { sa_assign : reg → expr reg → STm.t incr E unit
  ; sa_forget : reg → STm.t incr E unit
  ; sa_store  : expr reg → expr reg → STm.t incr E unit
  ; sa_vload : reg → AST.memory_chunk → AST.ident → ptrofs → STm.t incr E var
  ; sa_vstore : AST.memory_chunk → AST.ident → ptrofs → reg → STm.t incr E var
  ; sa_annot_input : int → int → STm.t incr E unit
  ; sa_return : expr reg → STm.t incr E var
  ; sa_init : RTL.genv → mem → result E ab_state
  ; sa_eval : expr reg → STm.t incr E Values.val
  ; sa_assert : expr reg → STm.t incr E unit
  ; sa_is_int : ab_state → reg → bool
  }.

End STATE.

Inductive aevent var :=
| AEinput (κ: AST.memory_chunk) (s: AST.ident) (ofs: ptrofs) (v: var)
| AEannot (msg: String.string) (args: list int)
| AEoutput (κ: AST.memory_chunk) (s: AST.ident) (ofs: ptrofs) (v: var).

Arguments AEannot {var} msg args.

Definition atrace var := list (aevent var).

Section SPEC.

Import Values.

Context (ab_state: Type) (incr: relation ab_state).
Variables E var: Type.
Variable abState : state_abstraction incr E var.

Variable prog : RTL.program.
Let ge := Genv.globalenv prog.

Hypothesis solution : ab_state → (var → Values.val) → Prop.

Definition event_of_aevent ρ (aevt: aevent var) : event :=
  match aevt with
  | AEinput κ s ofs v =>
    Event_vload κ s ofs match ρ v with Values.Vint i => EVint i | _ => EVint Int.zero end
  | AEannot msg args =>
    Event_annot msg (List.map EVint args)
  | AEoutput κ s ofs v => Event_vstore κ s ofs match ρ v with Values.Vint i => EVint i | _ => EVint Int.zero end
  end.

Definition concrete_trace ρ : atrace var → trace :=
  List.map (event_of_aevent ρ).

(* Definition fmem : Type := (int + reg → Values.val)%type. *)
Definition match_istate : Type := (var → Values.val) → ab_state → Values.val → RTL.regset → mem → Prop.

Hypothesis mi: match_istate.

Record state_abstraction_spec : Prop :=
{
sa_assign_sound : ∀ dst e s s' H,
  sa_assign abState dst e s = STm.Result tt s' H →
  ∀ ρ,
    solution s' ρ →
    ∀ rs m sp,
      mi ρ s sp rs m→
      ∃ v,
        rtl_eval_expr ge sp rs m e tt = STm.Result v tt I ∧
        mi ρ s' sp (rs # dst <- v) m
;
sa_forget_sound :
  ∀ {dst σ σ' Hincr},
    sa_forget abState dst σ = STm.Result tt σ' Hincr →
    ∀ {ρ},
      solution σ' ρ →
      ∀ {rs m sp},
        mi ρ σ sp rs m →
        mi ρ σ' sp (rs # dst <- Vundef) m
;
sa_store_sound :
  ∀ {dst e σ σ' Hincr},
    sa_store abState dst e σ = STm.Result tt σ' Hincr →
    ∀ {ρ},
      solution σ' ρ →
      ∀ {rs m sp},
        mi ρ σ sp rs m →
        ∀ v a m',
          rtl_eval_expr ge sp rs m e tt = STm.Result v tt I →
          rtl_eval_expr ge sp rs m dst tt = STm.Result a tt I →
          Mem.storev AST.Mint32 m a v = Some m' →
          mi ρ σ' sp rs m'
;
sa_vload_sound :
  ∀ {dst κ n ofs σ v σ' Hincr},
    sa_vload abState dst κ n ofs σ = STm.Result v σ' Hincr →
    κ = AST.Mint32 →
    ∀ {ρ},
      solution σ' ρ →
      ∀ {rs m sp},
        mi ρ σ sp rs m →
        mi ρ σ' sp (rs # dst <- (Val.load_result κ (ρ v))) m
      ∧ match ρ v with Vint _ => True | _ => False end
;
sa_vstore_sound :
  ∀ {κ n ofs src σ v σ' Hincr},
    sa_vstore abState κ n ofs src σ = STm.Result v σ' Hincr →
    κ = AST.Mint32 →
    ∀ {ρ},
      solution σ' ρ →
      ∀ {rs m sp},
        mi ρ σ sp rs m →
        ρ v = rs # src ∧
        mi ρ σ' sp rs m
      ∧ match ρ v with Vint _ => True | _ => False end
;
sa_annot_input_sound :
  ∀ {min max σ σ' Hincr},
    sa_annot_input abState min max σ = STm.Result tt σ' Hincr →
    ∀ {ρ},
      solution σ' ρ →
      ∀ {rs m sp},
        mi ρ σ sp rs m →
        mi ρ σ' sp rs m
;
sa_return_sound :
  ∀ {r σ} {v:var} {σ' H},
    sa_return abState (EVar r) σ = STm.Result v σ' H →
    ∀ {ρ}, solution σ ρ →
           ∀ {sp rs m},
             mi ρ σ sp rs m →
             ρ v = rs # r
      ∧ match ρ v with Vint _ => True | _ => False end
;
sa_init_sound m₀ σ₁:
  sa_init abState ge m₀ = Result σ₁ →
  ∀ ρ,
  solution σ₁ ρ →
  ∀ (m₁: mem) (stk b: block) fd,
    Genv.init_mem prog = Some m₀ →
    Mem.alloc m₀ 0 (RTL.fn_stacksize fd) = (m₁, stk) →
    Genv.find_symbol ge (AST.prog_main prog) = Some b →
    Genv.find_funct_ptr ge b = Some (AST.Internal fd) →
    mi ρ σ₁ (Vptr stk Ptrofs.zero) (RTL.init_regs nil (RTL.fn_params fd)) m₁
;
sa_eval_sound {e σ v σ' H} :
  sa_eval abState e σ = STm.Result v σ' H →
  ∀ ρ,
    solution σ' ρ →
    ∀ rs m sp,
      mi ρ σ sp rs m→
      rtl_eval_expr ge sp rs m e tt = STm.Result v tt I ∧
      mi ρ σ' sp rs m
;
 sa_assert_sound {e σ σ' H}:
   sa_assert abState e σ = STm.Result tt σ' H →
  ∀ ρ,
    solution σ' ρ →
    ∀ rs m sp,
      mi ρ σ sp rs m→
      rtl_eval_expr ge sp rs m e tt = STm.Result (Values.Vint Int.one) tt I ∧
      mi ρ σ' sp rs m
;
sa_is_int_sound : ∀ σ r,
    sa_is_int abState σ r = true →
    ∀ ρ,
      solution σ ρ →
    ∀ rs m sp,
      mi ρ σ sp rs m →
      match rs # r with Vint i => True | _ => False end
}.

End SPEC.
