Require Import Bool.
Require Import MoreLib.
Require Import ZArith Psatz Morphisms.

Set Implicit Arguments.
Unset Elimination Schemes.

Local Open Scope Z_scope.
Coercion is_true : bool >-> Sortclass.

Record interval := { imin : Z ; imax : Z }.

Definition denote (a: interval) (i: Z) : bool :=
  (imin a <=? i ) && (i <=? imax a).

Definition equiv (a b: interval) : bool :=
  (imax a <? imin a) && (imax b <? imin b)
  ||
  (imin a =? imin b) && (imax a =? imax b).

Definition eqb (a b: interval) : { a = b } + { a ≠ b }.
  refine (
    match Z.eq_dec (imin a) (imin b), Z.eq_dec (imax a) (imax b) with
    | left EQ, left EQ' => left _
    | _, _ => right _
    end);
  abstract (destruct a; destruct b; simpl in *; congruence).
Defined.

Definition leb (a b: interval) : bool :=
  (imin b <=? imin a) && (imax a <=? imax b).

Definition size (a: interval) : N :=
  match imax a - imin a with
  | Zpos p => Npos p
  | _ => N0
  end%Z.

Instance size_proper : Proper (equiv ==> eq) size.
Proof.
  unfold equiv, size, is_true.
  intros x x' Hx. rewrite orb_true_iff in Hx.
  destruct Hx as [h | h]; rewrite andb_true_iff in h; destruct h as (l & r).
  rewrite Z.ltb_lt in l, r.
  destruct (imax x - imin x) eqn: ?.
  destruct (imax x' - imin x') eqn: ?; reflexivity || lia. lia.
  destruct (imax x' - imin x') eqn: ?; reflexivity || lia.
  rewrite Z.eqb_eq in l, r. rewrite l, r. reflexivity.
Qed.

Module Notations.
Notation "i ∈ a" := (denote a i) (at level 60, no associativity) : interval_scope.
Notation "a == b" := (equiv a b) (at level 70, no associativity) : interval_scope.
(* Notation "a ⊆ b" := (leb a b) (at level 70, no associativity) : interval_scope. *)
Delimit Scope interval_scope with interval.
End Notations.
Import Notations.
Local Open Scope interval_scope.

Definition const (i: Z) : interval :=
  {| imin := i ; imax := i |}.

Definition bit : interval :=
  {| imin := 0; imax := 1 |}.

Definition is_const (a: interval) : option Z :=
  if imin a =? imax a then Some (imin a) else None.

Definition add (a b: interval) : interval :=
  {| imin := imin a + imin b ; imax := imax a + imax b |}.

Definition opp (a: interval) : interval :=
  {| imin := - imax a ; imax := - imin a |}.

Definition sub (a b: interval) : interval :=
  {| imin := imin a - imax b ; imax := imax a - imin b |}.

Definition muln (n: Z) (a: interval) : interval :=
  if 0 <=? n then {| imin := imin a * n ; imax := imax a * n |}
  else {| imin := imax a * n ; imax := imin a * n |}.

Definition mul (a b: interval) : interval :=
  let (am, aM) := a in let (bm, bM) := b in
  if 0 <=? am then
    if 0 <=? bm then {| imin := am * bm ; imax := aM * bM |}
    else if bM <=? 0 then {| imin := aM * bm ; imax := am * bM |}
         else {| imin := aM * bm ; imax := aM * bM |}
  else if aM <=? 0
       then if 0 <=? bm then {| imin := am * bM ; imax := aM * bm |}
            else if bM <=? 0 then {| imin := aM * bM ; imax := am * bm |}
                 else {| imin := am * bM ; imax := am * bm |}
       else if 0 <=? bm then {| imin := am * bM ; imax := aM * bM |}
            else if bM <=? 0 then {| imin := aM * bm ; imax := am * bm |}
                 else {| imin := Z.min (am * bM) (aM * bm) ; imax := Z.max (am * bm) (aM * bM) |}.

Require Import PP.PpsimplZ.

Lemma equiv_denote a b :
  a == b ↔ ∀ i, (i ∈ a ↔ i ∈ b).
Proof.
  unfold equiv, denote, is_true.
  split; intros H.
  intros; ppsimpl; lia.
  rewrite orb_true_iff.
  destruct ((_ <? _) && _) eqn:?. now left. right.
  generalize (proj1 (H (imax a))).
  generalize (proj1 (H (imin a))).
  generalize (proj2 (H (imax b))).
  generalize (proj2 (H (imin b))).
  clear H. ppsimpl. lia.
Qed.

Lemma leb_denote a b :
  leb a b →
  ∀ i, i ∈ a → i ∈ b.
Proof. unfold leb, denote, is_true. intros. ppsimpl. lia. Qed.

Lemma leb_trans b a c :
  leb a b → leb b c → leb a c.
Proof. unfold leb, denote, is_true. ppsimpl. lia. Qed.

Lemma const_correct i :
  i ∈ const i.
Proof. unfold const, denote, is_true. simpl. ppsimpl. lia. Qed.

Lemma const_exact i j :
  i ∈ const j → i = j.
Proof. unfold const, denote, is_true. simpl. ppsimpl. lia. Qed.

Corollary const_iff i j : i ∈ const j ↔ i = j.
Proof. split. apply const_exact. intros ->. apply const_correct. Qed.

Lemma is_const_elim a x :
  is_const a = Some x →
  x ∈ a ∧ ∀ i, i ∈ a → i = x.
Proof.
unfold is_const, denote, is_true. remember (imin a =? imax a) as b. intros H. destruct b. apply some_eq in H. split; intros; ppsimpl; Psatz.lia.
exact (some_neq (eq_sym H) _).
Qed.

Lemma size_const i : size (const i) = N0.
Proof. unfold size, const. simpl. now replace (Z.sub i i) with Z0 by Psatz.lia. Qed.

Lemma add_correct a b i j :
  i ∈ a → j ∈ b → i + j ∈ add a b.
Proof. unfold add, denote, is_true. simpl. ppsimpl. lia. Qed.

Lemma add_exact k a b :
  (∃ i, i ∈ a) → (∃ j, j ∈ b) →
  k ∈ add a b → ∃ i j, i ∈ a ∧ j ∈ b ∧ i + j = k.
Proof.
unfold add, denote, is_true. simpl.
intros (i & Hi) (j & Hj).
assert (k - imin b <= imax a ∨ imax a < k - imin b) as OR. lia. destruct OR.
exists (k - imin b), (imin b). ppsimpl. lia.
assert (k - imin a <= imax b ∨ imax b < k - imin a) as OR. lia. destruct OR.
exists (imin a), (k - imin a). ppsimpl. lia.
exists (k - imax b), (imax b). ppsimpl. lia.
Qed.

Lemma add_comm a b : add a b == add b a.
Proof. apply equiv_denote. intros i. unfold add, denote, is_true. ppsimpl. lia. Qed.

Lemma add_zero a : add a (const 0) == a.
Proof. apply equiv_denote. intros i. unfold add, const, denote, is_true. ppsimpl. lia. Qed.

Lemma sub_correct a b i j :
  i ∈ a → j ∈ b → i - j ∈ sub a b.
Proof. unfold sub, denote, is_true. simpl. ppsimpl. lia. Qed.

Lemma muln_correct n a i :
  i ∈ a → n * i ∈ muln n a.
Proof.
  unfold muln, denote, is_true.
  remember (0 <=? n) as b. destruct b; simpl;
  ppsimpl; nia.
Qed.

Lemma muln_exact n a i :
  i ∈ muln n a → n = 0 ∧ i = 0 ∨ n ≠ 0 ∧ i / n ∈ a.
Proof.
unfold muln, denote, is_true.
destruct (0 <=? n) eqn: ?;
ppsimpl; nia.
Qed.

Lemma muln_one n : muln n (const 1) == const n.
Proof.
  Opaque Z.mul.
  apply equiv_denote. intros i. unfold muln, const, denote, is_true. simpl.
  destruct (0 <=? n) eqn: ?; ppsimpl; lia.
  Transparent Z.mul.
Qed.

Lemma mul_correct a b i j :
  i ∈ a → j ∈ b → i * j ∈ mul a b.
Proof.
  unfold mul, denote, is_true.
  destruct a as (am & aM), b as (bm & bM). simpl.
  repeat match goal with |- context[ if ?b then _ else _ ] => let b' := fresh "b" in remember b as b'; destruct b'; simpl end;
  try (ppsimpl; nia).
  assert (i <= 0 ∨ 0 < i) by lia. ppsimpl. nia.
Qed.

Definition disjoint (a b: interval) : bool :=
  (imax a <? imin b) || (imax b <? imin a).

Lemma in_not_disjoint i a b :
  (i ∈ a) && (i ∈ b) && disjoint a b = false.
Proof.
  unfold denote, disjoint. ppsimpl. lia.
Qed.

Definition meet (a b: interval) : bool := negb (disjoint a b).
