Require Import Utf8 List Bool.
Require MoreMaps.
Require Import LinearCombination.
Import RField Ring_tac EquivDec BinPos.

Local Open Scope positive_scope.
Local Open Scope algebra_scope.

(** Implementation of linear combinations using lists. *)

Module LC : LinearCombinationType.
Definition var := positive.

Definition t f := list (var * f).
Definition zero : ∀ f, t f := λ f, nil.
Definition single {f} (v: var) (x: f) : t f := (v, x) :: nil.
Definition add {f feq} `{EqDec f feq} `{Plus f} `{Zero f} (a b: t f) : t f := rev_append a b.
Fixpoint rev_map f K (l tl: t f) : t f :=
  match l with nil => tl | (v, x) :: l' => rev_map f K l' ((v, K x) :: tl) end.
Definition sub {f feq} `{EqDec f feq} `{Plus f} `{Zero f} `{Opposite f} (a b: t f) : t f :=
  rev_map f (λ x, - x) b a.
Definition muln {f feq} `{EqDec f feq} `{Zero f} `{Mult f} (n: f) (a: t f) : t f :=
  if n ==b 0 then zero f else
  rev_map f (λ x, n × x) a nil.

Definition as_map {f feq} `{EqDec f feq} `{Plus f} `{Zero f} (m: t f) : Maps.PTree.t f :=
  List.fold_left (λ a (kv: var * f), let (k, v) := kv in let v' := match Maps.PTree.get k a with None => v | Some w => v + w end in if 0 == v' then Maps.PTree.remove k a else Maps.PTree.set k v' a) m (Maps.PTree.empty f).

Definition fold {f feq} `{EqDec f feq} `{Plus f} `{Zero f} A (K: A → var → f → A) (m: t f) (a: A) : A :=
  Maps.PTree.fold K (as_map m) a.

Definition is_zero {f feq} `{eq:EqDec f feq} `{Zero f} (a: t f) : bool :=
  List.forallb (λ x, snd x ==b 0) a.

Fixpoint is_single {f feq} `{eq:EqDec f feq} `{Zero f} (x: var) (a: t f) : option f :=
  match a with
  | nil => Some 0
  | y::a =>
    let (k, v) := y in
    if v ==b 0 then is_single x a
    else if (k =? x) && (is_zero a) then Some v
    else None
  end.

Definition den_aux {f} `{Plus f} `{Zero f} `{Mult f} (ω: var) (ρ: var → f) (a: t f) : f → f :=
  List.fold_left (λ r (kv: var * f), let (k,v) := kv in if k <=? ω then ρ k × v + r else r) a.

Definition den {f} `{Plus f} `{Zero f} `{Mult f} (ω: var) (ρ: var → f) (a: t f) : f :=
  den_aux ω ρ a 0.

Lemma den_zero f `{Field f} ω ρ :
  den ω ρ (zero f) === 0.
Proof. reflexivity. Qed.

Lemma den_single f `(Field f) ω ρ :
    ∀ x n, x <= ω → den ω ρ (single x n) === ρ x × n.
Proof. intros x n Hxω. unfold den, single. simpl. rewrite <- Pos.leb_le in Hxω. now rewrite Hxω, right_identity. Qed.

Remark den_aux_add f `(Field f) ω ρ :
  ∀ a z, den_aux ω ρ a z === den ω ρ a + z.
Proof.
  unfold den, den_aux.
  induction a as [ | (k, v) a IH ].
  simpl. intros z. now rewrite left_identity.
  simpl. intros z.
  rewrite ! (IH (if (_:bool) then (_ + _) else _)). clear IH.
  case (k <=? ω).
  - now rewrite associativity, right_identity.
  - now rewrite right_identity.
Qed.

Lemma den_add f `(Field f) ω ρ :
  ∀ a b: t f, den ω ρ (add a b) === den ω ρ a + den ω ρ b.
Proof.
  intros a b.
  unfold add.
  induction a as [ | (k, v) a IH ] using rev_ind.
  unfold den. simpl. now rewrite left_identity.
  rewrite rev_append_rev, rev_app_distr. simpl. rewrite <- rev_append_rev.
  unfold den, den_aux. rewrite fold_left_app. simpl.
  fold (den_aux ω ρ (rev_append a b)). rewrite (den_aux_add _ _).
  rewrite IH. clear IH.
  fold (den_aux ω ρ a) (den_aux ω ρ b) (den ω ρ a) (den ω ρ b).
  case (k <=? ω).
  - now rewrite right_identity, commutativity, associativity.
  - now rewrite right_identity.
Qed.

Section FIELD.
Context f `(F: Field f).
  
Add Field relevent : (field_of_field F).

Lemma den_sub ω ρ :
  ∀ a b: t f, den ω ρ (sub a b) === den ω ρ a − den ω ρ b.
Proof.
  intros a b. revert a.
  induction b as [ | (k, v) b IH ]; intros a.
  unfold den, sub. simpl. ring.
  unfold den, sub. simpl.
  rewrite (den_aux_add _ _ _ _ _ (if (_:bool) then (_ + _) else _)).
  fold (sub ((k, - v) :: a) b).
  fold (den ω ρ (sub ((k, - v) :: a) b)).
  rewrite (IH).
  unfold den at 1. simpl. case (k <=? ω).
  - rewrite (den_aux_add _ _). fold (den ω ρ a). rewrite !right_identity. ring.
  - ring.
Qed.

Lemma den_muln_aux ω ρ :
  ∀ (n: f) (a b: t f), den ω ρ (rev_map f (λ x, n × x) a b) === n × den ω ρ a + den ω ρ b.
Proof.
  intros n a.
  induction a as [ | (k, v) a IH ]; intros b; unfold den; simpl.
  ring.
  rewrite (den_aux_add _ _ _ _ _ (if (_:bool) then (_ + _) else _)).
  fold (den ω ρ (rev_map f (λ x, n × x) a ((k, n × v) :: b))).
  rewrite (IH).
  fold (den ω ρ b).
  unfold den at 2. simpl. case (k <=? ω).
  - rewrite (den_aux_add _ _).
    ring.
  - rewrite right_identity. reflexivity.
Qed.
  
Lemma den_muln ω ρ :
  ∀ (n: f) (a: t f), den ω ρ (muln n a) === n × den ω ρ a.
Proof.
  intros n a. unfold muln.
  unfold equiv_decb. case (equiv_dec n 0); intro EQ; simpl.
  - rewrite den_zero, EQ; eauto. ring.
  - rewrite (den_muln_aux ω ρ n a nil).
    change (den ω ρ nil) with 0. ring.
Qed.

Lemma den_is_zero ω ρ :
  ∀ a: t f, is_zero a = true → den ω ρ a === 0.
Proof.
  induction a as [ |[x c] a IHa].
  - reflexivity.
  - unfold den. simpl. rewrite andb_true_iff. intros [H6 H7]. apply IHa in H7. clear IHa. revert H6. unfold equiv_decb. case (equiv_dec c 0); try discriminate.
    intros H6 _. case (x <=? ω); auto.
    rewrite den_aux_add.
    * rewrite H6, H7. ring.
    * exact F.
Qed.

Lemma den_is_single ω ρ :
  ∀ (a: t f) (x: var) (v: f), x <= ω →
    is_single x a = Some v → den ω ρ a === ρ x × v.
Proof.
  induction a as [ |[k v] a IHa]; simpl; intros k' v' Hω.
  - intro H10. inversion H10; subst; clear H10.
    assert (H10: ρ k' × 0 === 0) by ring.
    rewrite H10. reflexivity.
  - unfold equiv_decb. case (equiv_dec v 0); intro EQ; simpl.
    * intro H10. unfold den. simpl. case (k <=? ω); auto.
      rewrite den_aux_add; eauto.
      rewrite IHa; eauto.
      rewrite EQ. ring.
    * case_eq (k =? k'); intro EQ2; simpl; try discriminate.
      case_eq (is_zero a); intro EQ3; simpl; try discriminate.
      intro H10; inversion H10; subst; clear H10.
      unfold den. simpl. apply Peqb_true_eq in EQ2. subst k'.
      rewrite <- Pos.leb_le in Hω. rewrite Hω.
      rewrite den_aux_add; eauto.
      rewrite den_is_zero; auto.
      ring.
Qed.


End FIELD.
End LC.
