Require Import Bool List.
Require Import MoreLib.
Require Import Relation_Definitions RelationClasses RelationPairs Morphisms.

Unset Elimination Schemes.
Set Implicit Arguments.

Definition option_bind A B (f: A → option B) (m: option A) : option B :=
  match m with
  | Some a => f a
  | None => None
  end.

Lemma option_bind_some A B (f: A → option B) m b :
  option_bind f m = Some b → ∃ a, m = Some a ∧ f a = Some b.
Proof.
unfold option_bind.
destruct m. eauto. intros H. exact (some_neq (eq_sym H) _).
Qed.

Module Result.
Inductive result (E R: Type) : Type :=
| Result (r: R)
| Error (e: E).
Global Arguments Result [E] [R] r.
Global Arguments Error [E] [R] e.

Lemma result_eq {E R} (r r': R) : @Result E R r = Result r' → r = r'.
Proof. refine (λ H, match H in _ = q return match q with Error _ => r = r' | Result k => _ = _ end with eq_refl => eq_refl end). Qed.

Lemma result_neq {E R} (e: E) (r': R) : Error e = Result r' → ∀ P : Prop, P.
Proof. refine (λ H, match H in _ = q return match q with Error _ => True | Result k => ∀ P : Prop, P end with eq_refl => I end). Qed.

Definition result_bind {E R R'} (f: R → result E R') (m: result E R) : result E R' :=
  match m with Result r => f r | Error e => Error e end.

Definition result_lift {E R R'} (f: R → R') (m: result E R) : result E R' :=
  result_bind (λ r, Result (f r)) m.

Lemma result_bind_inv {E R R'} (f: R → result E R') (m: result E R) (r: R') :
  result_bind f m = Result r →
  ∃ x, f x = Result r ∧ m = Result x.
Proof.
  unfold result_bind. destruct m . intros; eauto. exact (λ K, result_neq K _).
Qed.

Extraction Inline result_bind.

Module ResultNotations.
  Notation "a >>= b" := (result_bind b a) (at level 100).
  Notation "'do' x ← m ; q" := (m >>= λ x, q) (at level 100, x ident).
  Notation "'do' '_' ← m ; q" := (m >>= λ _, q) (at level 100).
End ResultNotations.
End Result.
Export Result.

(* A state and error monotone monad. *)
Module STm.

Section STATE.
Context (S : Type) `(incr: relation S)
        `{incr_refl: Reflexive S incr}
        `{incr_trans: Transitive S incr}.

Inductive result (E A: Type) (s: S) : Type :=
| Result (a: A) (s': S) `(incr s s')
| Error (e: E)
.
Arguments Result {E A} [s] a s' _.
Arguments Error {E A} [s] e.

Definition t (E A:Type) : Type := ∀ s, result E A s.

Definition ret {E A} (a: A) : t E A := λ s, Result a s (incr_refl _).
Definition err {E A} (e: E) : t E A := λ _, Error e.
Definition result_incr {E A} s s' (H: incr s s') (r: result E A s') : result E A s :=
  match r with Result a s'' H' => Result a s'' (incr_trans H H') | Error e => Error e end.
Definition bind {E A B} (f: A → t E B) (m: t E A) : t E B :=
  λ s, match m s with Result a s' H => result_incr H (f a s') | Error e => Error e end.

Definition lift {E A B} (f: A → B) : t E A → t E B := bind (λ a, ret (f a)).
Definition get {E} : t E S := λ s, ret s s.

Definition monotone A : Type := ∀ s, sig (incr s) * A.
Definition upd {E A} (f: monotone A) : t E A :=
  λ s, let '(exist s' H, a) := f s in Result a s' H.

Definition res {E A} (r: Result.result E A) : t E A :=
  match r with Result.Error e => err e | Result.Result a => ret a end.

Definition res' {E A} (r: S → Result.result E A) : t E A :=
  λ s, match r s with Result.Error e => err e s | Result.Result a => ret a s end.

Lemma result_eq E A s (a a': A) q q' (H: incr s q) (H': incr s q') :
    Result (E := E) a q H = Result a' q' H' → a = a' ∧ ∃ EQ: q = q', H' = eq_ind q (incr s) H q' EQ.
Proof.
  refine (λ EQ,
  match EQ in _ = r
    return match r with Error _ => False | Result a' q' H' => a = a' ∧ ∃ EQ: q = q', H' = eq_ind q (incr s) H q' EQ end
  with
  | eq_refl => conj eq_refl (ex_intro _ eq_refl eq_refl)
  end).
Qed.

Lemma result_neq E A s (e: E) (a: A) s' (H: incr s s') :
  Error e = Result a s' H → ∀ P : Prop, P.
Proof.
  refine (λ EQ, match EQ in _ = r return match r with Result _ _ _ => ∀ P, P | Error _ => True end with eq_refl => I end).
Qed.

Lemma bind_inv {E A B} m f s r s' Hincr:
  @bind E A B f m s = Result r s' Hincr →
  ∃ x s'' Hincr'' Hincr', f x s'' = Result r s' Hincr'' ∧ m s = Result x s'' Hincr' ∧ Hincr = eq_ind s' (incr s) (incr_trans Hincr' Hincr'') s' eq_refl.
Proof.
  unfold bind. destruct (m s) as [ x s'' Hincr' | ? ]; intros K.
  2: exact (result_neq K _).
  exists x, s''.
  destruct (f x s'') as [ y s''' Hincr'' | ? ].
  2: exact (result_neq K _).
  apply result_eq in K. destruct K as (Hy & Hs''' & HHincr). subst.
  eexists Hincr'', Hincr'. repeat split.
Qed.

Lemma res_inv {E A} (r: Result.result E A) s r' s' Hincr :
  res r s = Result r' s' Hincr →
  r = Result.Result r' ∧ s = s'.
Proof.
  unfold res. destruct r; intros K. apply result_eq in K. destruct K as (? & ? & _); now subst. exact (result_neq K _).
Qed.

Lemma res'_inv {E A} (r: S → Result.result E A) s r' s' Hincr :
  res' r s = Result r' s' Hincr →
  r s = Result.Result r' ∧ s = s'.
Proof.
  unfold res'. destruct (r s); intros K. apply result_eq in K. destruct K as (? & ? & ?). now subst. exact (result_neq K _).
Qed.

End STATE.

Module Notations.
  Notation "a >>= b" := (bind b a) (at level 100).
  Notation "'do' x ← m ; q" := (m >>= λ x, q) (at level 100, x ident).
  Notation "'do' '_' ← m ; q" := (m >>= λ _, q) (at level 100).
End Notations.

Section MORE.

Import Notations.

Context {S S': Type} (incr: relation S) (incr': relation S')
        `{incr_refl: Reflexive _ incr} `{incr'_refl: Reflexive _ incr'}
        `{incr_trans: Transitive _ incr}
        {E A: Type}.

Definition left (m: t incr E A) : t (incr * incr')%signature E A :=
  λ ss', let '(s, s') := ss' in match m s with Error e => Error _ _ _ e | Result a q H => Result (incr * incr')%signature E (s, s') a (q, s') (conj H (reflexivity _)) end.

Definition right (m: t incr' E A) : t (incr * incr')%signature E A :=
  λ ss', let '(s, s') := ss' in match m s' with Error e => Error _ _ _ e | Result a q H => Result (incr * incr')%signature E (s, s') a (s, q) (conj (reflexivity _) H) end.

Definition assert {P Q} (b: {P} + {Q} ) (e: unit → E) : t incr E unit :=
  if b then ret tt else err incr (e tt).

Definition map {B} (f: A → t incr E B) : list A → t incr E (list B) :=
  fix map (l: list A) {struct l} : t incr E (list B) :=
  match l with
  | nil => ret nil
  | a :: l' => do b ← f a ; do r ← map l'; ret (b :: r)
  end.

Definition iter (f: A → t incr E unit) : list A → t incr E unit :=
  fix iter (l: list A) {struct l} : t incr E unit :=
  match l with
  | nil => ret tt
  | a :: l' => do _ ← f a ; iter l'
  end.

Lemma result_incr_inv (s s': S) (Hincr: incr s s') (r: result incr E A s'):
  ∀ v s'' Hincr',
    result_incr Hincr r = Result _ _ _ v s'' Hincr' →
    ∃ Hincr'', r = Result _ _ _ v s'' Hincr''.
Proof.
  intros v s'' Hincr'. destruct r. intros K; apply result_eq in K. destruct K as (? & ? & _); subst; econstructor; eauto.
  exact (λ K, result_neq K _).
Qed.

Lemma left_inv (m: t incr E A) (ss': S * S') q q' r INCR:
  left m ss' = Result _ _ _ r (q, q') INCR → ∃ X, m (fst ss') = Result _ _ _ r q X ∧ q' = snd ss'.
Proof.
  destruct ss' as (s & s'). simpl. destruct (m s) as [ ? ? INCR' | ? ]; intros K.
  apply result_eq in K. destruct K as (? & K & _). apply pair_eq in K. destruct K; subst.
  exists INCR'; intuition congruence. exact (result_neq K _).
Qed.

Lemma right_inv (m: t incr' E A) (ss': S * S') q q' r INCR:
  right m ss' = Result _ _ _ r (q, q') INCR → ∃ X, m (snd ss') = Result _ _ _ r q' X ∧ q = fst ss'.
Proof.
  destruct ss' as (s & s'). simpl. destruct (m s') as [ ? ? INCR' | ? ]; intros K.
  apply result_eq in K. destruct K as (? & K & _). apply pair_eq in K. destruct K; subst.
  exists INCR'; intuition congruence. exact (result_neq K _).
Qed.

End MORE.

End STm.
Arguments STm.Result {S} [incr] {E A} [s] _ s' _.

(* Trivially monotone relation. *)
Module U.
  Section S.
  Context (S: Type).
  Definition incr (_ _: S) := True.
  Instance incr_refl : Reflexive incr := λ _, I.
  Instance incr_trans : Transitive incr := λ _ _ _ _ _, I.
  Definition ret {E A} (a: A) : STm.t incr E A := STm.ret a.
  Definition err {E A} (e: E) : STm.t incr E A := STm.err incr e.
  Definition put {E} (s: S) : STm.t incr E unit := λ s₀, STm.Result tt s I.
  End S.
End U.
Arguments U.incr {S} _ _.

Lemma uincr_I {S} {u v: S} (H: U.incr u v) : H = I.
Proof. now destruct H. Qed.

Require Import MoreLib.
From Coq
Require Import String.
From compcert
Require Import Maps Integers.
From compcert
Require Errors.

Lemma int_eq_true {i j: int} : Int.eq i j = true → i = j.
Proof. intros H. generalize (Int.eq_spec i j). rewrite H. exact id. Qed.

Lemma int_eq_false {i j: int} : Int.eq i j = false → i ≠ j.
Proof. intros H. generalize (Int.eq_spec i j). rewrite H. exact id. Qed.

Require Import Eqdep_dec.

Lemma unit_dec : ∀ x y: unit, { x = y } + { x ≠ y }.
Proof. now intros () (); left. Qed.
Lemma True_dec : ∀ x y: True, { x = y } + { x ≠ y }.
Proof. now intros () (); left. Qed.

Ltac hsplit :=
  unfold U.ret, U.err in *;
  unfold STm.upd, STm.get, STm.ret, STm.err in *;
  unfold STm.lift, result_lift in *;
  repeat match goal with
  | H : _ ∧ _ |- _ => let H' := fresh H in destruct H as ( H & H' )
  | H : unit |- _ => destruct H
  | H : prod _ _ |- _ => let H' := fresh H in destruct H as ( H & H' )
  | H : ∃ a, _ |- _ => let n := fresh a in destruct H as ( n & H )
  | H : tt = tt |- _ => assert (H = eq_refl) by (apply (UIP_dec unit_dec)); subst
  | H : I = I |- _ => assert (H = eq_refl) by (apply (UIP_dec True_dec)); subst
  | H : option_bind _ _ = Some _ |- _ => apply option_bind_some in H
  | H : Some _ = Some _ |- _ => apply some_eq in H
  | H : Some _ = None |- _ => exact (some_neq H _)
  | H : None = Some _ |- _ => exact (some_neq (eq_sym H) _)
  | H : inl _ = inl _ |- _ => apply inl_eq in H
  | H : inr _ = inr _ |- _ => apply inr_eq in H
  | H :  _ :: _ = _ :: _ |- _ => apply cons_eq in H
  | H : _ && _ = true |- _ => apply andb_true_iff in H
  | H : Int.eq _ _ = true |- _ => apply int_eq_true in H
  | H : Int.eq _ _ = false |- _ => apply int_eq_false in H
  | H : U.incr _ _ |- _ => pose proof (uincr_I H); try subst H
    (*
  | H : _ = Result _ |- _ =>
    first [ apply ST.bind_inv in H | apply ST.right_inv in H | apply ST.left_inv in H
            | apply ptree_at_inv in H | apply result_eq in H | exact (result_neq H _) | apply ST.res_inv in H
            | apply result_bind_inv in H ]
            *)
  | H : STm.left _ _ = _ |- _ => pose proof (STm.left_inv _ H); clear H
  | H : STm.right _ _ = _ |- _ => pose proof (STm.right_inv _ H); clear H
  | H : STm.assert ?b _ _ = STm.Result _ _ _ |- _ => (destruct b; [ apply STm.result_eq in H | exact (STm.result_neq H _) ])
  | H : STm.result_incr _ _ = STm.Result _ _ _ |- _ => apply STm.result_incr_inv in H
  | H : Result _ = Result _ |- _ => apply result_eq in H
  | H : Error _ = Result _ |- _ => exact (result_neq H _)
  | H : result_bind _ _ = Result _ |- _ => apply result_bind_inv in H
  | H : (_, _) = (_, _) |- _ => apply pair_eq in H
  | H : STm.res _ _ = STm.Result _ _ _ |- _ => apply STm.res_inv in H
  | H : STm.res' _ _ = STm.Result _ _ _ |- _ => apply STm.res'_inv in H
  | H : STm.Result _ _ _ = STm.Result _ _ _ |- _ => apply STm.result_eq in H
  | H : STm.Error _ _ _ _ = STm.Result _ _ _ |- _ => exact (STm.result_neq H _)
  | H : STm.bind _ _ _ = STm.Result _ _ _ |- _ => apply STm.bind_inv in H
  end.

Ltac flatten n :=
  repeat match goal with H : context[ match ?a with _ => _ end] |- _ => destruct a; n end.

Ltac flatten_eq n :=
  repeat match goal with H : context[ match ?a with _ => _ end] |- _ => destruct a eqn: ?; n end.
