Require Export Utf8.
Require Import EquivDec.
Require Psatz.
Import Bool List ZArith.

Unset Elimination Schemes.
Set Implicit Arguments.

(* Some extra stuff that should belong to stdlib. *)

Lemma and_forall_comm A (P Q:A → Prop) :
  (∀ x, P x ∧ Q x) <-> ((∀ x, P x) ∧ (∀ x, Q x)).
Proof. firstorder. Qed.

Lemma NPeano_leb_lt x y : Nat.leb x y = false ↔ (y < x)%nat.
Proof.
  rewrite <- not_true_iff_false.
  rewrite (PeanoNat.Nat.leb_le x y).
  exact (PeanoNat.Nat.nle_gt x y).
Qed.

Section ASSOC.

Variables A B : Type.
Hypothesis beq : A → A → bool.

Variable a : A.
Fixpoint assoc (l: list (A * B)) : option B :=
  match l with nil => None | (a', b) :: l' => if beq a' a then Some b else assoc l' end.

End ASSOC.

Lemma Forall2_nil_inv A B f (l: list B):
  Forall2 f (@nil A) l → l = nil.
Proof. refine (λ H, match H in Forall2 _ u v return match u return Prop with nil => v = nil | _ => True end with Forall2_nil _ => eq_refl | _ => I end). Qed.

Lemma Forall2_cons_inv_l A B f (a: A) l l':
  Forall2 f (a :: l) l' →
  ∃ (b: B) q, l' = b :: q ∧ f a b ∧ Forall2 f l q.
Proof.
  exact (λ H, match H in Forall2 _ u v return match u return Prop with a :: l => ∃ b q, v = b :: q ∧ f a b ∧ Forall2 f l q | _ => True end with Forall2_nil _ => I | Forall2_cons x y X IH => ex_intro _ _ (ex_intro _ _ (conj eq_refl (conj X IH))) end).
Qed.

Definition some_eq {A} (a a': A) : Some a = Some a' → a = a'.
Proof. exact (λ H, match H in _ = b return match b with Some a' => a = a' | None => False end with eq_refl => eq_refl end). Qed.

Definition some_neq {A} (a: A) : Some a = None → ∀ P : Prop, P.
Proof. exact (λ H, match H in _ = b return match b with Some _ => True | None => _ end with eq_refl => I end). Qed.

Lemma pair_eq {A B} (a a': A) (b b': B) : (a, b) = (a', b') → a = a' ∧ b = b'.
Proof. refine (λ H, match H in _ = q return let '(x, y) := q in a = x ∧ b = y with eq_refl => conj eq_refl eq_refl end). Qed.

Lemma inl_eq {A B} (a a': A) : inl B a = inl B a' → a = a'.
Proof. refine (λ H, match H in _ = q return match q with inr _ => a = a' | inl k => _ = _ end with eq_refl => eq_refl end). Qed.

Lemma inr_eq {A B} (b b': B) : inr A b = inr A b' → b = b'.
Proof. refine (λ H, match H in _ = q return match q with inr b' => b = b' | inl _ => b = b' end with eq_refl => eq_refl end). Qed.

Definition cons_eq {A} (a a': A) (q q': list A) : a :: q = a' :: q' → a = a' ∧ q = q'.
Proof. exact (λ H, match H in _ = b return match b with a' :: q' => a = a' ∧ q = q' | nil => False end with eq_refl => conj eq_refl eq_refl end). Qed.

Definition is_nil {A} (l: list A) : { l = nil } + { l ≠ nil}.
refine
  match l with nil => left eq_refl | _ => right _ end.
Proof.
abstract (
  intros K;
  refine match K in _ = q return match q with nil => False | _ => True end with eq_refl => I end).
Defined.

Definition sum_beq A B (a_beq: A → A → bool) (b_beq: B → B → bool) : A + B → A + B → bool :=
  λ x x',
  match x, x' with
  | inl a, inl a' => a_beq a a'
  | inr b, inr b' => b_beq b b'
  | _, _ => false
  end.

Definition prod_beq A B (a_beq: A → A → bool) (b_beq: B → B → bool) : A * B → A * B → bool :=
  λ ab ab', let '(a, b) := ab in let '(a', b') := ab' in a_beq a a' && b_beq b b'.

Instance PosEqDec : EqDec positive (@eq positive) := Pos.eq_dec.

Lemma if_dec_true P A (b: {P} + {¬P}) (x y: A): P → (if b then x else y) = x.
Proof. now destruct b. Qed.
Lemma if_dec_false P A (b: {P} + {¬P}) (x y: A): ¬ P → (if b then x else y) = y.
Proof. now destruct b. Qed.

Local Open Scope Z_scope.

Definition Zzerop (z: Z) : { z = 0 } + { z ≠ 0 } :=
  match z as z' return { z' = 0 } + { z' ≠ 0 } with
  | Z0 => left eq_refl
  | Zpos p => right (λ K : Zpos p = 0, match K in _ = k return match k with Zpos _ => True | _ => False end with eq_refl => I end)
  | Zneg _ => right (λ K : Zneg _ = 0, match K in _ = k return match k with Zneg _ => True | _ => False end with eq_refl => I end)
  end.

Definition not_none A (m: option A) : m ≠ None → { a: A | m = Some a } :=
  match m with
  | Some a => λ _, exist _ a eq_refl
  | None => λ K, False_rect _ (K eq_refl)
  end.

(* Notations about sets *)
Notation "P ⊆ Q" := (∀ a, P a → Q a) (at level 70).
Notation "P ∩ Q" := (λ a, P a ∧ Q a) (at level 19).
Notation "P ∪ Q" := (λ a, P a ∨ Q a) (at level 19).
Notation "f ∘ g" := (λ x, f (g x)) (at level 40, left associativity).

Require Psatz.
Import NArith ZArith.

Lemma N_div_eucl_full (a b: N) :
  (let '(q, r) := N.div_eucl a b in a = b * q + r ∧ (b = 0 ∧ r = a ∨ 0 <= r < b))%N.
Proof.
destruct a as [ | a].  simpl; Psatz.lia.
destruct b as [ | b].  simpl; Psatz.lia.
assert (Npos b ≠ N0) as NZ by Psatz.lia.
generalize (N.pos_div_eucl_remainder a (Npos b) NZ).
generalize (N.pos_div_eucl_spec a (Npos b)).
simpl.
destruct (N.pos_div_eucl a (Npos b)).
intros ->. simpl. intros Hr.
destruct n; Psatz.lia.
Qed.

Ltac elim_div :=
  unfold N.div, N.modulo, Z.div, Z.modulo in *;
  repeat match goal with
  | H : context [ N.div_eucl ?X ?Y ] |- _ =>
    generalize (N_div_eucl_full X Y); destruct (N.div_eucl X Y)
  | |- context [ N.div_eucl ?X ?Y ] =>
    generalize (N_div_eucl_full X Y); destruct (N.div_eucl X Y)
  | H : context[ Z.div_eucl ?X ?Y ] |- _ =>
    generalize (Z_div_mod_full X Y) ; destruct (Z.div_eucl X Y)
  | |- context[ Z.div_eucl ?X ?Y ] =>
    generalize (Z_div_mod_full X Y) ; destruct (Z.div_eucl X Y)
  end; unfold Remainder.

From compcert
Require Import Integers.

Lemma unsigned_inj i j : Ptrofs.unsigned i = Ptrofs.unsigned j → i = j.
Proof. destruct i, j; apply Ptrofs.mkint_eq. Qed.

Lemma add_inj i j k : Ptrofs.add i j = Ptrofs.add i k → j = k.
Proof.
  intros H.
  cut (Ptrofs.add (Ptrofs.add i (Ptrofs.neg i)) j = Ptrofs.add (Ptrofs.add i (Ptrofs.neg i)) k).
  now rewrite Ptrofs.add_neg_zero, !Ptrofs.add_zero_l.
  rewrite (Ptrofs.add_commut i), !Ptrofs.add_assoc. congruence.
Qed.

Lemma add_repr x y : Int.add (Int.repr x) (Int.repr y) = Int.repr (x + y).
Proof.
  apply Int.eqm_samerepr, Int.eqm_add; apply Int.eqm_unsigned_repr_l, Int.eqm_refl.
Qed.

Lemma sub_repr x y : Int.sub (Int.repr x) (Int.repr y) = Int.repr (x - y).
Proof.
  apply Int.eqm_samerepr, Int.eqm_sub; apply Int.eqm_unsigned_repr_l, Int.eqm_refl.
Qed.

Lemma mul_repr x y : Int.mul (Int.repr x) (Int.repr y) = Int.repr (x * y).
Proof.
  apply Int.eqm_samerepr, Int.eqm_mult; apply Int.eqm_unsigned_repr_l, Int.eqm_refl.
Qed.

Lemma repr_inj a b :
  - Int.max_unsigned <= b - a <= Int.max_unsigned →
  a ≠ b →
  Int.repr a ≠ Int.repr b.
Proof.
  intros R NE EQ.
  apply (f_equal Int.unsigned) in EQ. rewrite ! Int.unsigned_repr_eq in EQ.
  unfold Int.max_unsigned in R.
  assert (0 < b - a ∨ b - a < 0) as OR by Psatz.lia.
  destruct OR.
  cut ((b - a) mod Int.modulus = 0).
  rewrite Z.mod_small; Psatz.lia.
  rewrite Zdiv.Zminus_mod, Zeq_minus; easy.
  cut ((a - b) mod Int.modulus = 0).
  rewrite Z.mod_small; Psatz.lia.
  rewrite Zdiv.Zminus_mod, Zeq_minus; easy.
Qed.

Lemma repr_inj' (a b: Z) : a mod Int.modulus ≠ b mod Int.modulus → Int.repr a ≠ Int.repr b.
Proof.
  intros NE EQ.
  apply (f_equal Int.unsigned) in EQ. rewrite ! Int.unsigned_repr_eq in EQ.
  exact (NE EQ).
Qed.

Lemma test_sign_bit (n: Z) :
  Int.half_modulus <= n < Int.modulus →
  Z.shiftr n 31 = 1.
Proof.
  change Int.modulus with 4294967296.
  change Int.half_modulus with 2147483648.
  intros H. rewrite Z.shiftr_div_pow2. 2: discriminate.
  change (2 ^ 31) with 2147483648.
  elim_div. Psatz.lia.
Qed.
