Require Import Utf8.
Require Run RTLCompiler.
Require RunProof RTLComplements.
Require Import QapGen.
Require QAPproof AbStateImplProof.

Import ZArith RelationClasses Morphisms.
Import Maps Integers.
Import MoreLib Monads LinearCombination modp modpProof AbState AbStateImpl QAP.
Require Import project.
Import LargePrime.

(** This hypothesis is discharged in primality.v and uses Théry et alii CoqPrime. *)
Axiom large_prime_is_prime : Znumtheory.prime (Zpos large_prime).

Notation "⟦ p ⟧" := (Behaviors.program_behaves (Csem.semantics p)).

(** Good programs are those that have no wrong behaviours. *)
Definition good (prog: Csyntax.program) : Prop :=
  ⟦ prog ⟧ ⊆ Behaviors.not_wrong.

Section ORACLE.

Context
  (O: Type)
  (oracle_gen: AbStateImpl.oracle positive O)
  (oracle_eval: @AbInt.Arith.t unit Errors.errmsg (modp.zp large_prime) U.incr → AbStateImpl.oracle (modp.zp large_prime) O)
.

Existing Instances QAP.incr_refl QAP.incr_trans U.incr_refl U.incr_trans.

Variable prog : Csyntax.program.
Hypothesis good_program : good prog.
Let ge := Csem.genv_genv (Csem.globalenv prog).

Let F := modp_field _ large_prime_is_prime.
Existing Instances zpEquiv zpPlus zpZero zpMult zpOne.
Let var := positive.

Definition lift_solution {var} (σ: sa_state var) (ρ: var → Z) : val var → Values.val :=
  λ x,
  match x with
  | Const c => Expr.eval_constant ge σ.(sa_sp) c
  | Var x => Values.Vint (Int.repr (ρ x))
  end.

Definition is_io {t} (q: cqap t) (w: var) : bool :=
  match (meta_data q) ! w
  with
  | Some (Bank_Shared r) => true
  | _ => false
  end.

Definition extends {t} (q: cqap t) (ρio: var → Z) (ρ: var → Z) : Prop :=
  ∀ w, is_io q w →ρio w = ρ w.

Definition extends' {t} (σ: sa_state var) (q: cqap t) (ρio: var → Z) (ρ': val var → Values.val) : Prop :=
  ∃ ρ, extends q ρio ρ ∧ ρ' = lift_solution σ ρ.

Definition concrete_behaviour {var} (ρ: val var → Values.val) (tr: list (aevent (val var))) (v: int)
  : Behaviors.program_behavior :=
  Behaviors.Terminates (concrete_trace ρ tr) v.

Theorem qap_gen_correct tc cse cp dc fuel tr v σ :
  qap_gen O oracle_gen tc cse cp dc prog fuel = Result (tr, v, σ) →
  let q : cqap _ := snd σ in
  let σ : sa_state var := fst (fst σ) in
  ∀ ρio,
  (∃ ρ, extends q ρio ρ ∧
    (QAPproof.cqap_solution (f_of_z:=zconst _)) project q ρ) →
  ∃ ρ',
  extends' σ q ρio ρ' ∧
    ∃ i,
    ρ' v = Values.Vint i ∧
     ⟦ prog ⟧ (concrete_behaviour ρ' tr i).
Proof.
  destruct σ as ((q & o) & σ). simpl.
  intros H ρio (ρ & EXT & H').
  set (ρ' := lift_solution q ρ).
  exists ρ'. split. exists ρ. split. exact EXT. reflexivity.
  unfold qap_gen in H.
  destruct (RTLCompiler.transf_c_program tc cse cp dc prog) as [rtl | msg] eqn: hRTL. 2: hsplit.
  pose proof (RTLComplements.transf_c_program_is_refinement tc cse cp dc prog rtl hRTL good_program) as hRefinement.
  cut (∃ i, ρ' v = Values.Vint i ∧ Behaviors.program_behaves (RTL.semantics rtl) (Behaviors.Terminates (concrete_trace ρ' tr) i)).
  intros (i & Hv & beh). exists i. split. exact Hv. apply hRefinement. exact beh.
  refine (RunProof.simulation rtl _ _ _ _ _ _ _
    (AbStateImplProof.sa_sound
      _
       (QAPproof.cqap_monotone project)
       (QAPproof.cQAP_sound F _ (refl_equal _) project_range _ project_const eq_refl eq_refl _ project_add project_sub project_muln project_mul _ project_inj project_subrange)
       rtl)
    _
    (λ beh H, good_program beh (hRefinement beh H))
    fuel tr v _ H _ _).
  - clear.
  unfold AbStateImplProof.ms.
  intros rs rs' H ρ [ [σ σ'] σ'' ] sp m [ A B ].
  refine (conj _ B).
  intros x v H0. rewrite (H x). auto.
- clear. intros x H.
  apply not_none in H. destruct H as (x' & H).
  pose proof (project_range _ _ H) as Hc.
  apply Interval.const_exact in Hc. subst.
  destruct (project_eqm _ _ H) as (k & Hmod).
  destruct x as (x & X). apply modp.zp_eq. simpl.
  apply modp.overflow_elim. intros z [ (_ & <-) | K ].
  2: exfalso; clear -K; unfold large_prime in K; Psatz.lia.
  clear H. simpl in *.
  cut (k = Z0). intros ->. Psatz.lia.
  apply ltb_lt in X. apply N2Z.inj_lt in X.
  unfold large_prime in *. simpl in *. zify. Psatz.lia.
- clear. intros x xr xr' a a' H1 H2 H3.
  refine (project_fun x xr xr' H1 H2 (Interval.leb_trans _ _ _ H3 _)). exact eq_refl.
- clear. intros σ σ'0 ((K & ?) & ?) a (Hc & ρ'' & Hsol). hsplit. hnf in K. simpl in *. split.
  + simpl. intros c. now rewrite Hc, K.
  + exists ρ''. split. simpl. eapply QAPproof.cqap_monotone; eauto. easy.
- split. subst ρ'. simpl. subst ge.
  intros [ ? | ? | s ofs]; only 1-2: reflexivity.
  unfold Expr.eval_constant.
  now rewrite (RTLCompiler.symbols_preserved tc cse cp dc prog rtl s hRTL).
  exists ρ. split. assumption. easy.
Qed.

Theorem eval_correct tc cse cp dc fuel tr' v' σ' :
  QapGen.eval O oracle_eval tc cse cp dc prog fuel = Result (tr', v', σ') →
  let ρ' := lift_solution (fst (fst σ')) (λ x, Z.of_N (modp.zp_to_n x)) in
  ∃ v,
    ρ' v' = Values.Vint v ∧
    Behaviors.program_behaves (Csem.semantics prog)
     (Behaviors.Terminates (concrete_trace ρ' tr') v).
Proof.
intros H ρ'.
unfold QapGen.eval in H.
destruct (RTLCompiler.transf_c_program tc cse cp dc prog) as [rtl | msg] eqn: hRTL. 2: hsplit.
pose proof (RTLComplements.transf_c_program_is_refinement tc cse cp dc prog rtl hRTL good_program) as hRefinement.
cut (∃ v, ρ' v' = Values.Vint v ∧ Behaviors.program_behaves (RTL.semantics rtl) (Behaviors.Terminates (concrete_trace ρ' tr') v)).
intros (v & Hv & beh). exists v. split. exact Hv. exact (hRefinement _ beh).
assert (H100: (Int.max_unsigned < Z.pos large_prime)%Z) by reflexivity.
refine (RunProof.simulation rtl _ _ _ _ _ _ _
   (AbStateImplProof.sa_sound
      _
     _
     (AbIntModp.modp_arith_sound H100)
      rtl)
   _
   (λ beh H, good_program beh (hRefinement beh H))
   fuel tr' v' σ' H _ _).
  - clear.
  unfold AbStateImplProof.ms.
  intros rs rs' H ρ [ [σ σ'] σ'' ] sp m [ A B ].
  refine (conj _ B).
  intros x v H0. rewrite (H x). auto.
- exact (λ _ _ _ _, id).
- intros σ σ'0 ((K & ?) & ?) a (Hc & ρ'' & Hsol). hsplit. hnf in K. simpl in *. split. simpl. intros c. now rewrite Hc, K.
  exists ρ''. now split.
- split. subst ρ'; simpl. subst ge.
  intros [ ? | ? | s ofs]; only 1-2: reflexivity.
  unfold Expr.eval_constant.
  now rewrite (RTLCompiler.symbols_preserved tc cse cp dc prog rtl s hRTL).
  exists (Z.of_N ∘ @modp.zp_to_n large_prime). split; reflexivity.
Qed.

End ORACLE.

Check qap_gen_correct.
Print Assumptions qap_gen_correct.
