Require Import Utf8 String.
Require Import MoreLib MoreString.

Import Relation_Definitions RelationClasses RelationPairs Morphisms.
Import Coqlib Maps Op Registers RTL Values Integers Globalenvs.

Require Import Monads AbState.
Import Expr.

Unset Elimination Schemes.
Set Implicit Arguments.
Unset Strict Implicit.

Section S.

Let E : Type := Errors.errmsg.

Context (ab_state: Type) (sa_incr: relation ab_state) `{sa_incr_refl: Reflexive _ sa_incr} `{sa_incr_trans: Transitive _ sa_incr}.
Variables var: Type.
Variable abState : state_abstraction sa_incr E var.

(* Global state of the interpreter.
 The first component is the current program-point or the final value.
*)
Definition state := ((node + var) * ab_state)%type.

Definition incr : relation state := (U.incr * sa_incr)%signature.

Inductive outcome :=
| Ostep (tr: atrace var)
| Oend.
Definition tau_step : outcome := Ostep nil.
Definition visible_step ev : outcome := Ostep (ev :: nil).

Lemma ostep_eq tr tr' : Ostep tr = Ostep tr' → tr = tr'.
Proof. refine (λ H, match H in _ = q return match q with Oend => tr = tr' | Ostep tr' => tr = tr' end with eq_refl => eq_refl end). Qed.

Lemma ostep_neq tr' : Oend = Ostep tr' → ∀ P : Prop, P.
Proof. refine (λ H, match H in _ = q return match q with Oend => True | _ => _ end with eq_refl => I end). Qed.

Import STm.Notations.
Existing Instances U.incr_refl U.incr_trans.

Definition get_pc : STm.t incr E node :=
  STm.left (STm.res' (λ pc_r, match pc_r with inl pc => Result pc | inr _ => Error (Errors.msg "execution stopped") end)).

Definition set_pc' {E} (pc: node) : STm.t incr E unit :=
  STm.left (U.put (inl pc)).
Let set_pc E pc := do _ ← @set_pc' E pc; STm.ret tau_step.
Arguments set_pc {E} pc _.

Definition lea (addr: addressing) (args: list reg) : result E (expr reg) :=
  match addr, args with
  | Aindexed ofs, x :: nil => Result (EBinop Badd (EVar x) (EConst (ConstInt (Int.repr ofs))))
  | Aindexed2 ofs, x :: x' :: nil => Result (EBinop Badd (EBinop Badd (EVar x) (EVar x')) (EConst (ConstInt (Int.repr ofs))))
  | Ascaled sc ofs, x :: nil => Result (EBinop Badd (EBinop Bmul (EVar x) (EConst (ConstInt (Int.repr sc)))) (EConst (ConstInt (Int.repr ofs))))
  | Aindexed2scaled sc ofs, x :: x' :: nil => Result (EBinop Badd (EVar x) (EBinop Badd (EBinop Bmul (EVar x') (EConst (ConstInt (Int.repr sc)))) (EConst (ConstInt (Int.repr ofs)))))
  | Aglobal s ofs, nil => Result (EConst (ConstGlobal s ofs))
  | Abased s ofs, x :: nil => Result (EBinop Badd (EConst (ConstGlobal s ofs)) (EVar x))
  | Abasedscaled sc s ofs, x :: nil => Result (EBinop Badd (EConst (ConstGlobal s ofs)) (EBinop Bmul (EVar x) (EConst (ConstInt (Int.repr sc)))))
  | Ainstack ofs, nil => Result (EConst (ConstStackAddr ofs))
  | _, _ => Error (Errors.msg ("lea: unknown addressing " ++ string_of_addressing addr))
  end.

Definition expr_of_cond (cond: condition) (args: list reg) : result E (expr reg) :=
  match cond, args with
  | Ccomp c, x :: x' :: nil => Result (EBinop (Bcmp true c) (EVar x) (EVar x'))
  | Ccompu c, x :: x' :: nil => Result (EBinop (Bcmp false c) (EVar x) (EVar x'))
  | Ccompimm c i, x :: nil => Result (EBinop (Bcmp true c) (EVar x) (EConst (ConstInt i)))
  | Ccompuimm c i, x :: nil => Result (EBinop (Bcmp false c) (EVar x) (EConst (ConstInt i)))
  | _, _ => Error (Errors.msg ("bad cond"))
  end.

Definition expr_of_op (op: operation) (args: list reg) : result E (expr reg) :=
  match op, args with
  | Omove, src :: nil => Result (EVar src)
  | Ointconst i, nil => Result (EConst (ConstInt i))
  | Onot, x :: nil => Result (EBinop Bnot (EVar x) (EConst (ConstInt Int.zero)))
  | Osub, x :: x' :: nil => Result (EBinop Bsub (EVar x) (EVar x'))
  | Omul, x :: x' :: nil => Result (EBinop Bmul (EVar x) (EVar x'))
  | Omulimm i, x :: nil => Result (EBinop Bmul (EVar x) (EConst (ConstInt i)))
  | Oand, x :: x' :: nil => Result (EBinop Band (EVar x) (EVar x'))
  | Oandimm i, x :: nil => Result (EBinop Band (EVar x) (EConst (ConstInt i)))
  | Oor, x :: x' :: nil => Result (EBinop Bor (EVar x) (EVar x'))
  | Oorimm i, x :: nil => Result (EBinop Bor (EVar x) (EConst (ConstInt i)))
  | Oxor, x :: x' :: nil => Result (EBinop Bxor (EVar x) (EVar x'))
  | Oxorimm i, x :: nil => Result (EBinop Bxor (EVar x) (EConst (ConstInt i)))
  | Oshl, x :: x' :: nil => Result (EBinop Bshl (EVar x) (EVar x'))
  | Oshlimm i, x :: nil =>
    if Int.ltu i Int.iwordsize
    then Result (EBinop Bmul (EVar x) (EConst (ConstInt (Int.shl Int.one i))))
    else Error (Errors.msg "shlimm too large")
  | Oshru, x :: x' :: nil => Result (EBinop (Bshr false) (EVar x) (EVar x'))
  | Oshruimm i, x :: nil => Result (EBinop (Bshr false) (EVar x) (EConst (ConstInt i)))
  | Oshrimm i, x :: nil => Result (EBinop (Bshr true) (EVar x) (EConst (ConstInt i)))
  | Olea addr, _ => lea addr args
  | Ocmp cond, _ => expr_of_cond cond args
  | _, _ => Error (Errors.msg ("Unknown operator " ++ string_of_operator op))
  end.

Definition never_fails b : bool :=
  match b with Bshl | Bshr _ => false | _ => true end.

Fixpoint is_int p (e: expr reg) : bool :=
  match e with
  | EVar v => p v
  | EConst (ConstInt _) => true
  | EConst (ConstStackAddr _ | ConstGlobal _ _) => false
  | ELoad _ => false
  | EBinop op a b => never_fails op && is_int p a && is_int p b
  end.

Definition assert_is_int (e: expr reg) : STm.t incr E unit :=
  λ s,
  if is_int (abState.(sa_is_int) (snd s)) e
  then STm.ret tt s
  else STm.err _ (Errors.msg ("select: " ++ string_of_expr e ++ " maybe not an int")) s.

Section PROG.

Variable prog : program.
Let ge := Genv.globalenv prog.

Definition offset {A} (x: A) (i: nat) : expr A :=
  EBinop Badd (EVar x) (EConst (ConstInt (Int.repr (4 * Z.of_nat i)))).

Definition vload dst κ f ofs : STm.t incr E (option (aevent var)) :=
      do b ← match Genv.find_symbol ge f with Some b => STm.ret b | None => STm.err _ (Errors.MSG "bad volatile: " :: Errors.CTX f :: nil) end;
      do _ ← if Senv.block_is_volatile ge b then STm.ret tt else STm.err _ (Errors.msg "volatile access to non volatile");
      do _ ← match κ with AST.Mint32 => STm.ret tt | _ => STm.err _ (Errors.msg "wrong volatile type") end;
      do x ← STm.right (abState.(sa_vload) dst κ f ofs);
      STm.ret (Some (AEinput κ f ofs x)).

Definition step (c: code) : STm.t incr E outcome :=
  do pc ← get_pc;
  do i ← match c ! pc with Some i => STm.ret i | None => STm.err _ (Errors.MSG ("No instruction at ") :: Errors.POS pc :: nil) end;
  match i with
  | Inop pc' => set_pc pc'
  | Iop op args dst pc' =>
    do e ← STm.res (expr_of_op op args);
    do _ ← STm.right (abState.(sa_assign) dst e);
    set_pc pc'
  (*
  | Icall sg (inr fn) args dst pc' =>
    do ev ← match find_function ge (inr fn) (Regmap.init Vundef) with
              | Some (AST.External (AST.EF_external f sg')) =>
                do _ ← ST.assert (AST.signature_eq sg sg') (λ _, Errors.msg "signature mismatch");
                ST.right (abState.(sa_extcall) dst f (List.map EVar args))
              | Some (AST.External ef) => ST.err (Errors.MSG "unknown external call: " :: error_of_external ef )
              | Some (AST.Internal _) => ST.err (Errors.msg "function call")
              | None => ST.err (Errors.msg "bad function pointer")
              end;
      do _ ← set_pc pc';
      ST.ret (visible_step ev)
  *)
  | Ireturn (Some src) =>
    do v ← STm.right (abState.(sa_return) (EVar src));
    do _ ← STm.left (U.put (inr v));
    STm.ret Oend
  | Iload AST.Mint32 addr args dst pc' =>
    do a ← STm.res (lea addr args);
    do _ ← STm.right (abState.(sa_assign) dst (ELoad a));
    set_pc pc'
  | Istore AST.Mint32 addr args src pc' =>
    do a ← STm.res (lea addr args);
    do _ ← STm.right (abState.(sa_store) a (EVar src));
    set_pc pc'
  | Ibuiltin ef args dst pc' =>
    do ev ← match ef with
    | AST.EF_vload κ =>
      do dst ← match dst with AST.BR d => STm.ret d | _ => STm.err _ (Errors.msg "bad destination for volatile load") end;
      match args with
      | AST.BA src :: nil =>
        do addr ← STm.right (abState.(sa_eval) (EVar src));
        match addr with
        | Vptr b ofs =>
          match Genv.invert_symbol ge b with
          | Some i => vload dst κ i ofs
          | _ => STm.err _ (Errors.msg "vload: not a global")
          end
        | _ => STm.err _ (Errors.msg "vload: not a pointer")
        end
      | AST.BA_addrglobal f ofs :: nil =>
        vload dst κ f ofs
      | _ => STm.err _ (Errors.msg "vload: bad arguments")
      end
    | AST.EF_vstore κ =>
      do _ ← match dst with AST.BR_none => STm.ret tt | _ => STm.err _ (Errors.msg "bad destination for volatile store") end;
      match args with
      | AST.BA_addrglobal f ofs :: AST.BA arg :: nil =>
        do b ← match Genv.find_symbol ge f with Some b => STm.ret b | None => STm.err _ (Errors.MSG "bad volatile: " :: Errors.CTX f :: nil) end;
        do _ ← if Senv.block_is_volatile ge b then STm.ret tt else STm.err _ (Errors.msg "volatile write to non volatile");
        do _ ← match κ with AST.Mint32 => STm.ret tt | _ => STm.err _ (Errors.msg "wrong volatile type") end;
        do x ← STm.right (abState.(sa_vstore) κ f ofs arg);
        STm.ret (Some (AEoutput κ f ofs x))
      | _ => STm.err _ (Errors.msg "vstore: wrong arguments")
      end
    | AST.EF_annot _ annot (AST.Tint :: AST.Tint :: nil) =>
      do _ ← match dst with AST.BR_none => STm.ret tt | _ => STm.err _ (Errors.msg "bad destination for annotation") end;
      match args with
      | AST.BA_int low :: AST.BA_int hi :: nil =>
        do _ ← STm.assert (string_dec annot "input")
           (λ _, Errors.MSG "unknown annotation" :: Errors.MSG annot :: nil);
        do _ ← STm.right (abState.(sa_annot_input) low hi);
        STm.ret (Some (AEannot annot (low :: hi :: nil)))
      | _ => STm.err _ (Errors.msg "annot: wrong argumenst")
      end
    | _ => STm.err _ (Errors.MSG "unknown builtin: " :: error_of_external ef)
    end;
    do _ ← set_pc' pc';
    STm.ret match ev with Some e => visible_step e | None => Ostep nil end

  | Icond cond args if_so if_not =>
    do e ← STm.res (expr_of_cond cond args);
    (* Is this an assert? *)
    if match c ! if_not with
       | Some (Inop pc') =>
           match c ! pc' with Some (Inop loop) => Pos.eqb pc' loop | _ => false end
       | _ => false end
    then
      do _ ← STm.right (sa_assert abState e);
      set_pc if_so
    else
    (* Is this a select? *)
    match match c ! if_so with
       | Some (Iop op args' dst pc') =>
         if Pos.eqb pc' if_not
         then Some (op, args', dst)
         else None
       | _ => None end
    with Some (op, args', dst) =>
      do e' ← STm.res (expr_of_op op args');
      let select := EBinop Badd (EBinop Bmul e' e) (EBinop Bmul (EVar dst) (EBinop Bsub (EConst (ConstInt Int.one)) e)) in
      do _ ← assert_is_int e';
      do _ ← assert_is_int (EVar dst);
      do _ ← STm.right (abState.(sa_assign) dst select);
      set_pc if_not
    | None =>
    do b ← STm.right (sa_eval abState e);
    match b with
    | Vint i => if Int.eq i Int.one then set_pc if_so else if Int.eq i Int.zero then set_pc if_not else STm.err _ (Errors.msg ("bad boolean"))
    | _ => STm.err _ (Errors.msg ("very bad boolean"))
    end
    end

  | _ => STm.err _ (Errors.msg ("step: " ++ string_of_instruction i))
  end.

End PROG.

(* Terminating behavior: trace, final value, and final state of the interpreter. *)
Definition behavior := (atrace var * var * ab_state)%type.

Import ResultNotations.

Section MAIN_LOOP.
Variables (prog: program) (c : code).
Fixpoint main (fuel: nat) {struct fuel}:  STm.t (incr * @U.incr (atrace var))%signature E unit :=
  match fuel with
  | O => STm.err _ (Errors.msg "dame más gasolina")
  | S fuel' =>
    STm.bind
    (λ o,
     match o with
     | Oend => STm.ret tt
     | Ostep tr => STm.bind (λ _, main fuel') (STm.right (STm.bind (λ acc, U.put (rev_append tr acc)) (STm.get (incr := _))))
     end)
    (STm.left (step prog c))
  end.
End MAIN_LOOP.

Definition run (prog: program) (fuel: nat) : result E behavior :=
  let ge := Genv.globalenv prog in
  do m₀ ← match Genv.init_mem prog with Some m₀ => Result m₀ | None => Error (Errors.msg "init_mem failed") end;
  do b ← match Genv.find_symbol ge prog.(AST.prog_main) with Some b => Result b | None => Error (Errors.msg "find_symbol failed") end;
  do f ← match Genv.find_funct_ptr ge b with Some f => Result f | None => Error (Errors.msg "find_funct_ptr failed") end;
  do _ ← if AST.signature_eq (funsig f) AST.signature_main then Result tt else Error (Errors.msg "main has incorrect signature");
  do fd ← match f with AST.External _ => Error (Errors.msg "main is extern") | AST.Internal fd => Result fd end;
  let c : code := fd.(fn_code) in
  do i ← abState.(sa_init) ge m₀;
  match main prog c fuel ((inl fd.(fn_entrypoint), i), nil) with
  | STm.Error e => Error e
  | STm.Result tt (q, σ, tr) _ =>
    do v ← match q with inl pc => Error (Errors.msg "assert false") | inr v => Result v end;
    Result (rev' tr, v, σ)
  end.

End S.

Arguments Ostep {var} _.
Arguments Oend {var}.

Ltac ostep_inv :=
  repeat match goal with
  | H : _ = Ostep _ |- _ => apply ostep_eq in H || exact (ostep_neq H _)
  | H : _ = Oend |- _ => exact (ostep_neq (eq_sym H) _)
  end.
