Require Import Bool ZArith Znumtheory NArith Psatz.
Require Import Eqdep_dec.
Require Import MoreLib.

Set Implicit Arguments.
Local Unset Elimination Schemes.

Local Open Scope Z_scope.
Local Open Scope N_scope.

Coercion is_true : bool >-> Sortclass.

(** We implement the extended version of Euclid's algorithm,
    i.e. the one computing Bezout's coefficients as it computes
    the [gcd]. We follow the algorithm given in Knuth's
    "Art of Computer Programming", vol 2, page 325. *)

Section extended_euclid_algorithm.

  Import Znumtheory.

  Lemma Zis_gcd_for_euclid :
    forall a b d q:Z, Zis_gcd b (a - q * b) d -> Zis_gcd a b d.
  Proof.
    intros a b d q H. destruct H as [(u & ?) (v & ?) H'].
    constructor.
    exists (u * q + v). nia.
    exists u. assumption.
    intros x (y & Y) K. apply H'. exact K. destruct K as (k & K). exists (y - q * k). nia.
  Defined.

  Variables a b : Z.
  Hypotheses (a_pos : 0 <= a) (b_pos: 0 <= b).

  (** The specification of Euclid's algorithm is the existence of
      [u], [v] and [d] such that [ua+vb=d] and [(gcd a b d)]. *)

  Record eucl :=
    { u : Z ; v : Z ; d : Z ; bezout_eq: u * a + v * b = d ; gcd_pos : 0 <= d ; gcd : Zis_gcd a b d }.

  (** The recursive part of Euclid's algorithm uses well-founded
      recursion of non-negative integers. It maintains 6 integers
      [u1,u2,u3,v1,v2,v3] such that the following invariant holds:
      [u1*a+u2*b=u3] and [v1*a+v2*b=v3] and [gcd(u3,v3)=gcd(a,b)].
      *)

  Lemma euclid_rec :
    forall v3:Z,
      0 <= v3 ->
     ∀ u3,
    0 <= u3 ->
      forall u1 u2 v1 v2:Z,
	u1 * a + u2 * b = u3 ->
	v1 * a + v2 * b = v3 ->
	(forall d:Z, Zis_gcd u3 v3 d -> Zis_gcd a b d) -> eucl.
  Proof.
    refine (Zlt_0_rec _ _).
    intros x IH Hx_pos u3 H1 u1 u2 v1 v2 H2 H3 H4.
    destruct (Zzerop x) as [ Hxz | Hxnz ].
    - clear IH.
      refine {| u := u1 ; v := u2 ; d := u3 |}; try lia.
      apply H4; clear H4. rewrite Hxz. constructor. exists 1; lia. exists 0. exact eq_refl. exact (λ _ H _, H).
    - set (q := u3 / x).
      assert (Hq : 0 <= u3 - q * x < x) by (subst q; elim_div; lia).
      refine (IH _ Hq x Hx_pos v1 v2 (u1 - q * v1) (u2 - q * v2) _ _ _).
      lia. nia.
      intros; apply H4.
      apply Zis_gcd_for_euclid with q; assumption.
  Defined.

  (** We get Euclid's algorithm by applying [euclid_rec] on
      [1,0,a,0,1,b] when [b>=0] and [1,0,a,0,-1,-b] when [b<0]. *)

  Definition bezout : eucl.
  refine (match Z_le_gt_dec 0 b with
  | left LE => euclid_rec LE a_pos 1 0 0 1 _ _ (λ _, id)
  | right GT => @euclid_rec (- b) _ a a_pos 1 0 0 (-1) _ _ _
  end).
  Proof.
  abstract lia.
  abstract lia.
  lia.
  abstract lia.
  abstract lia.
  abstract lia.
  Defined.

End extended_euclid_algorithm.

Definition pos_is_pos (p: positive) : (0 <= Zpos p)%Z :=
 λ EQ : Lt = Gt, match EQ in _ = k return match k with Lt => True | _ => False end with eq_refl => I end.
Arguments pos_is_pos : clear implicits.
Definition zero_is_pos : (0 <= 0)%Z :=
 λ EQ : Eq = Gt, match EQ in _ = k return match k with Eq => True | _ => False end with eq_refl => I end.

Section MODP.

Variable p : positive.
Notation q := (Npos p).

Record zp : Type := Zp
 { zp_to_n :> N ; zp_range : zp_to_n <? q }.

Lemma zp_eq (x y: zp) :
  @eq N x y → x = y.
Proof.
  destruct x as (x & H), y as (y & K). simpl.
  intros <-. f_equal.
  apply UIP_dec, bool_dec.
Qed.

Lemma ltb_lt {x y: N} : N.ltb x y → N.lt x y.
Proof. exact (proj1 (N.ltb_lt _ _)). Qed.

Definition zp_dec (x y: zp) : { x = y } + { x ≠ y } :=
match N.eq_dec x y with | left EQ => left (zp_eq _ _ EQ) | right NE => right (λ K, NE (f_equal zp_to_n K)) end.

Definition zero : zp := Zp 0 eq_refl.
Definition mone : zp. refine {| zp_to_n := Npos p - 1 |}.
Proof. abstract (apply N.ltb_lt; lia). Defined.

Definition ltb x y : { r | r + y = x } + { x <? y }.
  refine(
    match Z.of_N y - Z.of_N x as r return _ = r → _ with
    | Zpos z => λ K, inright _
    | Z0 => λ K, inleft (exist _ N0 _)
    | Zneg z => λ K, inleft (exist _ (Npos z) _)
    end eq_refl)%Z.
Proof.
  abstract lia. abstract (apply N.ltb_lt; lia). abstract lia.
Defined.

Definition overflow (n: N) : N :=
  match ltb n (Npos p) with inleft (exist r _) => r | inright _ => n end.

Lemma overflow_elim' n (P: N → Prop) :
  (n < Npos p → P n) →
  (∀ x, x + Npos p = n → P x) →
  P (overflow n).
Proof.
  intros LT GE. unfold overflow. destruct ltb as [ (r & EQ) | LT' ]. auto. apply N.ltb_lt in LT'. auto.
Qed.

Lemma overflow_elim n (P: N → Prop) :
  (∀ z, n < Npos p ∧ n = z ∨ z + Npos p = n → P z) →
  P (overflow n).
Proof.
  intros H. apply overflow_elim'; intuition.
Qed.

Definition one : zp.
refine {| zp_to_n := overflow 1 |}.
Proof.
  abstract (apply overflow_elim; intuition; [ subst; now apply N.ltb_lt | replace z with 0 by lia ; exact eq_refl ]).
Defined.

Definition add (x y: zp) : zp.
  refine
  {| zp_to_n := overflow (x + y) |}.
Proof.
  abstract (
  destruct x as [x X], y as [y Y]; simpl; apply N.ltb_lt in X; apply N.ltb_lt in Y;
  apply N.ltb_lt;
  apply overflow_elim; intros; lia).
Defined.

Definition neg (x: zp) : zp.
refine
 match x : N with
  | N0 => x
  | Npos y => {| zp_to_n := Npos p - Npos y |}
  end.
Proof.
  abstract (apply N.ltb_lt; lia).
Defined.

Lemma neg_elim (x: zp) (P: N → Prop) :
  ((x : N) = 0 → P 0) →
  (0 < x < Npos p → P (Npos p - x)) →
  P (neg x).
Proof.
Opaque N.sub.
destruct x as [x Hx]. simpl.
destruct x as [ | x ]. exact (λ H _, H eq_refl).
intros _ H. apply H. apply N.ltb_lt in Hx. lia.
Transparent N.sub.
Qed.

Definition sub (x y: zp) : zp := add x (neg y).

Lemma pos_not_zero x : Npos x ≠ 0%N.
Proof.
  exact (λ H, match H in _ = k return if k then False else True with eq_refl => I end).
Qed.
Arguments pos_not_zero : clear implicits.

Lemma mod_ltb x y : (x mod Npos y <? Npos y = true)%N.
Proof.
  exact (proj2 (N.ltb_lt (x mod Npos y) (Npos y)) (N.mod_lt x (Npos y) (pos_not_zero y))).
Qed.

Fixpoint mod_mul_pn (x: positive) (y: N) : N :=
  match x with
  | 1 => y
  | p~0 => let z := mod_mul_pn p y in overflow (z + z)
  | p~1 => let z := mod_mul_pn p y in overflow (y + overflow (z + z))
  end%positive.

Definition mod_mul (x y: N) : N :=
  match x with N0 => x | Npos p => mod_mul_pn p y end.

Lemma mul2 (x: positive) (y: N) :
  Npos (x~0) * y = Npos x * y + Npos x * y.
Proof. lia. Qed.

Lemma mul2' (x: positive) (y: N) :
  Npos (x~1) * y = Npos x * y + Npos x * y + y.
Proof. lia. Qed.

Lemma mod_mul_spec' x y :
  x < Npos p →
  y < Npos p →
  mod_mul x y = (x * y) mod Npos p.
Proof.
Opaque N.mul.
  intros X Y.
  destruct x as [ | x ]. reflexivity.
  simpl mod_mul.
  revert x X y Y.
  induction x as [ x IH | x IH | ]; cycle 1.
  2: intros X [ | y ] Y; [ reflexivity | symmetry; apply N.mod_small, Y ].
  intros X y Y. assert (Npos x < Npos p) as X' by (clear -X; lia); clear X; specialize (IH X' y Y).
  simpl mod_mul_pn. rewrite IH; clear IH. rewrite mul2.
  assert (∀ a b q r q' r', a * b = Npos p * q + r → a * b + a * b = Npos p * q' + r' → r + r < Npos p → r' < Npos p → q' = q + q ∧ r' = r + r) as L.
  {
    intros a b q r q' r' A B R R'. rewrite A in B. clear a b A.
    assert (Npos p * (q + q) + (r + r) = Npos p * q' + r') as K.
    lia. clear B.
    cut (q + q = q').
    intros; subst. lia.
    nia.
  }
  apply overflow_elim. intuition subst.
  elim_div. simpl in *. intuition subst; try lia. symmetry. eapply L; eassumption.
  elim_div. simpl in *. intuition subst; try lia.
  rewrite H in H2. clear H x y X' Y.
  cut (Npos p * (n + n + 1) + z = Npos p * n1 + n2). 2: lia.
  intros.
  assert (z < Npos p). lia.
  cut (n + n + 1 = n1). 2: nia.
  intros <-. lia.
  intros X'.
  assert (Npos x < Npos p) as X by lia. clear X'.
  specialize (IH X).
  intros y Y. specialize (IH y Y).
  simpl. rewrite IH. clear IH.
  repeat apply overflow_elim.
  intuition subst; elim_div; intuition subst; try lia;
  simpl in *.
  rewrite mul2' in H3. rewrite H1 in H3. clear H1.
  assert (Npos p * (n + n) + (n0 + n0 + y) = Npos p * n1 + n2) by lia.
  assert (n + n = n1) by nia. subst. lia.
  rewrite mul2' in H3. rewrite H0 in H3. clear H0.
  assert (Npos p * (n + n + 1) + z0 = Npos p * n1 + n2) by lia.
  assert (n + n + 1 = n1) by nia. subst. lia.
  rewrite mul2' in H3. rewrite H1 in H3. clear H1.
  assert (Npos p * (n + n + 1) + (y + z) = Npos p * n1 + n2) by lia.
  assert (n + n + 1 = n1) by nia. subst. lia.
  rewrite mul2' in H3. rewrite H in H3. clear H.
  assert (Npos p * (n + n + 2) + z0 = Npos p * n1 + n2) by lia.
  assert (n + n + 2 = n1) by nia. subst. lia.
Qed.

Lemma mod_mul_spec (x y: zp) :
  mod_mul x y = (x * y) mod Npos p.
Proof.
  apply mod_mul_spec'; apply N.ltb_lt, zp_range.
Qed.
Global Opaque mod_mul.

Definition mul (x y: zp) : zp.
refine
  {| zp_to_n := mod_mul x y |}.
Proof.
  abstract (
  rewrite (mod_mul_spec x y);
  apply mod_ltb).
Defined.

Lemma add_comm (x y: zp) : add x y = add y x.
Proof.
  apply zp_eq. simpl. now rewrite N.add_comm.
Qed.

Lemma add_zero_l x : add zero x = x.
Proof.
  destruct x as [x X].
  apply zp_eq. simpl.
  apply ltb_lt in X.
  apply overflow_elim. intros. lia.
Qed.

Lemma add_zero_r x : add x zero = x.
Proof. rewrite add_comm. apply add_zero_l. Qed.

Lemma add_assoc (x y z: zp) : add (add x y) z = add x (add y z).
Proof.
  apply zp_eq. simpl.
  destruct x as [x X], y as [y Y], z as [q Q]; simpl in *; apply N.ltb_lt in X; apply N.ltb_lt in Y; apply N.ltb_lt in Q.
  repeat (apply overflow_elim); intros; lia.
Qed.

Lemma neg_neg (x: zp) : neg (neg x) = x.
Proof.
apply zp_eq. apply neg_elim; apply neg_elim; lia.
Qed.

Lemma add_neg x y : add (neg x) (neg y) = neg (add x y).
Proof.
Opaque N.sub.
destruct x as [x Hx], y as [y Hy].
apply zp_eq. simpl.
repeat match goal with
| |- context[ neg ] => apply neg_elim; simpl
| |- context[ overflow ] => apply overflow_elim
end; intros; lia.
Transparent N.sub.
Qed.

Lemma add_neg_r x : add x (neg x) = zero.
Proof.
apply zp_eq. simpl.
apply neg_elim. intros ->. reflexivity.
apply overflow_elim. intros z [(? & <-) | ?]; lia.
Qed.

Lemma mul_comm (x y: zp) : mul x y = mul y x.
Proof.
  apply zp_eq. simpl. rewrite ! mod_mul_spec. now rewrite N.mul_comm.
  
Qed.

Lemma mul_assoc (x y z: zp) : mul (mul x y) z = mul x (mul y z).
Proof.
  pose proof (pos_not_zero p) as NZ.
  apply zp_eq. simpl.
  rewrite ! mod_mul_spec, ! mod_mul_spec', N.mul_mod_idemp_r, N.mul_mod_idemp_l, N.mul_assoc; auto;
  apply N.ltb_lt; apply zp_range || apply mod_ltb.
Qed.

Lemma add_mul_distr_l x y z :
  mul x (add y z) = add (mul x y) (mul x z).
Proof.
  pose proof (pos_not_zero p) as NZ.
  apply zp_eq. simpl.
  rewrite ! mod_mul_spec.
  destruct x as [x X], y as [y Y], z as [q Q]; simpl in *; apply N.ltb_lt in X; apply N.ltb_lt in Y; apply N.ltb_lt in Q.
  rewrite mod_mul_spec'; auto.
  repeat apply overflow_elim.
  intros z [ (A & <- ) | A ] z' B.
  rewrite <- (N.mod_small _ _ A), <- N.add_mod. 2: easy. clear A.
  destruct B as [( B & <-) | B ]. rewrite Nmult_plus_distr_l; exact eq_refl.
  rewrite <- Nmult_plus_distr_l, <- B, Nmult_plus_distr_l.
  symmetry. apply N.mod_add. exact NZ.
  pose proof (λ x, N.mod_lt x _ NZ) as K. pose proof (K (x * y)). pose proof (K (x * q)). clear K.
  destruct B as [( B & <-) | B ].
  rewrite Nmult_plus_distr_l. rewrite N.add_mod, <- A. 2: exact NZ.
  change (Npos p) with (1 * Npos p) at 1. rewrite N.mod_add. 2: exact NZ.
  apply N.mod_small. lia.
  transitivity ((x * (z' + Npos p)) mod Npos p).
  rewrite Nmult_plus_distr_l, N.mod_add. exact eq_refl. exact NZ.
  rewrite B, Nmult_plus_distr_l, N.add_mod, <- A. 2: exact NZ.
  change (Npos p) with (1 * Npos p) at 1. rewrite N.mod_add. 2: exact NZ.
  apply N.mod_small. lia.
  apply overflow_elim; intuition lia.
Qed.

Lemma add_mul_distr_r x y z :
  mul (add x y) z = add (mul x z) (mul y z).
Proof.
  repeat rewrite (mul_comm _ z).
  apply add_mul_distr_l.
Qed.

Lemma mul_one_l x : mul one x = x.
Proof.
  destruct x as [x X].
  apply zp_eq. simpl.
  apply ltb_lt in X.
  rewrite mod_mul_spec'; auto.
  apply overflow_elim. intuition. subst. replace (1*x) with x by lia. now apply N.mod_small.
  assert (x = 0 ∧ p = xH ∧ z = 0). lia. intuition subst. reflexivity.
  apply overflow_elim. intuition lia.
Qed.

Lemma mul_neg_l x y : mul (neg x) y = neg (mul x y).
Proof.
Opaque N.sub.
pose proof (pos_not_zero p) as NZ.
destruct x as [x Hx], y as [y Hy].
pose proof (ltb_lt Hx) as X.
pose proof (ltb_lt Hy) as Y.
apply zp_eq.
unfold mul. simpl.
rewrite mod_mul_spec'; auto.
apply neg_elim; simpl. now intros ->.
apply neg_elim; simpl.
intros Hy' Hx'.
rewrite mod_mul_spec' in Hy'; auto.
rewrite (N.mod_divides _ _ NZ) in Hy'. destruct Hy' as (k & Hk).
rewrite N.mul_sub_distr_r, Hk.
rewrite (N.mod_divides _ _ NZ). exists (y - k). nia.
rewrite N.mul_sub_distr_r.
intros Hy' Hx'.
rewrite mod_mul_spec'; auto.
rewrite mod_mul_spec' in Hy'; auto.
elim_div. unfold fst, snd in *.
intros [ ? [ ? | ?] ]. lia.
intros [ ? [ ? | ?] ]. lia.
rewrite H1 in H.
assert (n1 = y - n - 1). nia.
subst. nia.
apply neg_elim; simpl; lia.
Transparent N.sub.
Qed.

Lemma mul_one_r x : mul x one = x.
Proof. rewrite mul_comm. apply mul_one_l. Qed.

Lemma mul_mone_l x : mul mone x = neg x.
Proof.
pose proof (pos_not_zero p) as NZ.
destruct x as [x Hx].
apply zp_eq. apply neg_elim. unfold mul, zp_to_n. intros ->. rewrite mod_mul_spec'. now rewrite N.mul_0_r.
unfold mone. lia. lia.
Opaque N.sub.
simpl. clear Hx. intros Hx.
rewrite mod_mul_spec'; try lia.
replace ((Npos p - 1) * x) with ((Npos p - x) + (x - 1) * Npos p) by nia.
rewrite N.mod_add; auto.
apply N.mod_small. lia.
Transparent N.sub.
Qed.

Lemma mul_mone_r x : mul x mone = neg x.
Proof. rewrite mul_comm. apply mul_mone_l. Qed.

Import Znumtheory.

Definition inverse (x: zp) : zp.
refine (
    let e := bezout (N2Z.is_nonneg x) (pos_is_pos p) in
    Zp (Z.to_N (e.(u) mod (Zpos p))) _).
Proof.
  abstract (
  apply N.ltb_lt;
  change (Npos p) with (Z.to_N (Zpos p)); rewrite <- Z2N.inj_lt;
    try (apply Z_mod_lt; exact eq_refl);
  refine (λ H : _ = Gt, match H in _ = k return match k with Lt => True | _ => False end with eq_refl => I end)).
Defined.

Definition const (n: N) : zp :=
  {| zp_to_n := n mod Npos p ; zp_range := mod_ltb _ _ |}.

Lemma const_inj_add n n' : const (n + n') = add (const n) (const n').
Proof.
pose proof (pos_not_zero p) as NZ.
apply zp_eq.
simpl. rewrite N.add_mod; auto. apply overflow_elim. intros z [ (H & <- ) | H ].
now apply N.mod_small.
rewrite <- H.
rewrite <- (N.mul_1_l (Npos p)) at 1. rewrite N.mod_add; auto. rewrite N.mod_small; auto.
generalize (N.mod_lt n' _ NZ).
generalize (N.mod_lt n _ NZ).
lia.
Qed.

Definition zconst (z: Z) : zp :=
  match z with
  | Z0 => zero
  | Zpos p => const (Npos p)
  | Zneg p => neg (const (Npos p))
  end.

Lemma prime_N_divisors q :
  prime (Zpos q) →
  ∀ a, (Z.of_N a | Zpos q) → a = Npos q ∨ a = 1%N.
Proof.
  intros PRIME a DIV.
  generalize (prime_divisors _ PRIME _ DIV). clear.
  lia.
Qed.

Lemma inverse_correct : prime (Zpos p) → ∀ x : zp, Z.of_N x ≠ 0 → mul x (inverse x) = const 1.
Proof.
  intros PRIME x Hx.
  apply zp_eq. simpl.
  destruct x as (x & X). simpl in *. apply N.ltb_lt in X.
  rewrite mod_mul_spec'; auto.
  destruct bezout as [u v d D K [ Q Y L ] ]. simpl in *.
  apply Zdivide_le in Q; auto. 2: lia.
  rewrite <- (Z2N.id d K) in Y. apply (prime_N_divisors PRIME) in Y.
  assert (Z.of_N x < Zpos p). lia. clear X.
  destruct Y as [Y | Y]. assert (d = Zpos p). rewrite <- (Z2N.id d K). lia. lia.
  assert (d = 1) as Hd. rewrite <- (Z2N.id d K). lia. subst d.
  rewrite <- (N2Z.id x) at 1.
  rewrite <- (Z2N.inj_mul). 2: lia. 2: apply Z_mod_lt; lia.
  change (Npos p) with (Z.to_N (Zpos p)).
  setoid_rewrite Hd in L.
  rewrite <- Z2N.inj_mod. 3: lia.
  rewrite Zmult_mod_idemp_r.
  set (y := Z.of_N x). fold y in H, Hx, K, Q, Y, L, Hd.
  replace (y * u) with (1 + (- v) * Zpos p) by lia.
  rewrite Z_mod_plus_full.
  now rewrite Z2N.inj_mod.
  cut (0 <= u mod Zpos p). nia.
  apply Z_mod_lt; lia.
  assert (∀ P: N → Prop, ∀ u : Z, (u <= 0 → P N0) → (∀ p, u = Zpos p → P (Npos p)) → P (Z.to_N u)) as U.
  { intros P [ | u | u ] LT GT; (apply LT || apply GT); lia. }
  apply U. intros _. lia. intros y Y. apply N2Z.inj_lt. simpl.
  rewrite <- Y. clear Y.
  apply Z_mod_lt; lia.
Qed.

Definition shift_inverse (n: Z) (v: zp) : zp :=
  inverse (sub v (zconst (n * Integers.Int.modulus))).

(* The hypothesis implies that 1 < p. *)
Remark one_is_one : prime (Zpos p) → const 1 = one.
Proof.
  intros PRIME. apply zp_eq. simpl. apply overflow_elim. intros z [ (H & ?) | H ]. rewrite N.mod_small; easy.
  destruct PRIME. lia.
Qed.

Lemma one_not_zero : prime (Zpos p) → one ≠ zero.
Proof.
unfold one, zero.
intros [X _] K. apply (f_equal zp_to_n) in K. simpl in K. revert K.
apply overflow_elim. intros ? [(? & ?) | ?]; lia.
Qed.

Lemma integral : prime (Zpos p) → ∀ x y : zp, mul x y = zero → x = zero ∨ y = zero.
Proof.
  intros PRIME x y H.
  destruct (Zzerop (Z.of_N x)) as [hx | hx]. left. apply zp_eq. apply N2Z.inj. rewrite hx. reflexivity.
  assert (mul x (mul y (inverse y)) = zero) as K. rewrite <- mul_assoc, H. apply zp_eq. reflexivity.
  destruct (Zzerop (Z.of_N y)) as [hy | hy]. right. apply zp_eq. apply N2Z.inj. rewrite hy. reflexivity.
  rewrite (inverse_correct PRIME y hy), (one_is_one PRIME), mul_one_r in K. now left.
Qed.

End MODP.
