Require Import Utf8.
From mathcomp
Require Import ssreflect ssrfun ssrbool eqtype ssrnat choice fintype seq.
From mathcomp
Require Import div bigop prime ssralg zmodp.
Import GRing.Theory.
Import BinPos BinInt.
Require Import Psatz.
Require LinearCombinationSpec modpProof.

Set Implicit Arguments.
Unset Strict Implicit.
Unset Printing Implicit Defensive.


Section ssrpositive.

  Lemma eqpositiveP : Equality.axiom Pos.eqb.
  Proof.
    move=> n m; apply: (iffP idP) => [|<-]; last by elim n.
      by elim: n m => [n IHn|n IHn|] [m|m|] //= /IHn->.
  Qed.

  Canonical positive_eqMixin := EqMixin eqpositiveP.
  Canonical positive_eqType := Eval hnf in EqType positive positive_eqMixin.

End ssrpositive.


Section ssrZ.

  Section EqType.

    Lemma eqZP : Equality.axiom Z.eqb.
    Proof.
      move=> n m; apply: (iffP idP) => [|<-].
      - by case: n m => [|n|n] [|m|m] //= /eqpositiveP ->.
      - by case n => [|p|p] //=; apply/eqP.
    Qed.

    Canonical Z_eqMixin := EqMixin eqZP.
    Canonical Z_eqType := Eval hnf in EqType Z Z_eqMixin.

  End EqType.


  Section ChoiceType.

    Definition findZ P n :=
      let z := Z_of_nat n in
      if P z then Some z else
      if P (- z)%Z then Some (- z)%Z else
        None.


    Lemma findZ_correct (P: pred Z) n x : findZ P n = Some x → P x.
    Proof.
      rewrite/findZ.
      case e₁: (P (Z.of_nat n)); first by move => [<-].
      by case e₂: (P (- Z.of_nat n)%Z); first by move => [<-].
    Qed.


    Lemma findZ_complete (P: pred Z) :
      (∃ x, P x) → ∃ n : nat, findZ P n.
    Proof.
      rewrite/findZ; move => [x Hx]. exists (Z.abs_nat x). rewrite Znat.Zabs2Nat.id_abs. move: Hx. case: (Z.abs_or_opp_abs x) => [{1}->|{1}->] => -> //=.
      by case: (P (Z.abs x)).
    Qed.


    Lemma findZ_extensional (P Q: pred Z) :
      P =1 Q → findZ P =1 findZ Q.
    Proof.
      rewrite/findZ => H n. rewrite !H. by case: (Q _); [|case: (Q _)].
    Qed.


    Fact Z_choiceMixin : choiceMixin Z.
    Proof. exists findZ; [exact findZ_correct|exact findZ_complete|exact findZ_extensional]. Qed.

    Canonical Z_choiceType := Eval hnf in ChoiceType Z Z_choiceMixin.

  End ChoiceType.


  Section ZModule.

    Canonical Z_zmodMixin := ZmodMixin Z.add_assoc Z.add_comm Z.add_0_l Z.add_opp_diag_l.
    Canonical Z_zmodType := Eval hnf in ZmodType Z Z_zmodMixin.

  End ZModule.


  Section Ring.

    Definition mulA : associative Z.mul := Z.mul_assoc.
    Definition mul1x : left_id 1%Z Z.mul := Z.mul_1_l.
    Definition mulx1 : right_id 1%Z Z.mul := Z.mul_1_r.
    Definition mulDx : left_distributive Z.mul Z.add := Z.mul_add_distr_r.
    Definition mulxD : right_distributive Z.mul Z.add := Z.mul_add_distr_l.

    Canonical Z_ringMixin := RingMixin Z.mul_assoc Z.mul_1_l Z.mul_1_r Z.mul_add_distr_r Z.mul_add_distr_l (eq_refl true).
    Canonical Z_ringType := Eval hnf in RingType Z Z_ringMixin.

  End Ring.


  Section ComRing.

    Definition mulC : commutative Z.mul := Z.mul_comm.

    Canonical Z_comRingType := Eval hnf in ComRingType Z_ringType mulC.

  End ComRing.

End ssrZ.


Require Import modp.
Import NArith.


Section Zmodp.

  Lemma N2Nat_inj_lt : forall x y,
            (x < y)%N <-> (N.to_nat x < N.to_nat y)%coq_nat.
  Proof. move => [|x] [|y] //=; by lia. Qed.

  Variable p' : positive.
  Local Notation p := (p' + 2)%positive.

  Let q' := Pos.to_nat p'.
  Local Notation q := q'.+2.

  Lemma pq : Pos.to_nat p = q.
  Proof. by rewrite -addn2 -plusE -(Pos2Nat.inj_add _ 2%positive). Qed.

  Hypothesis primep : Znumtheory.prime (Zpos p).

  Lemma primeq : prime.prime q.
  Proof.
    move: primep. rewrite/prime => PRIME. apply/primeP. split.
    - apply/ltP. by destruct PRIME; rewrite -pq -(Pos2Nat.inj_lt 1%positive).
    - move => d Hd. assert (H:(Z.of_N (N.of_nat d) | Z.pos p)%Z).
      * move/dvdnP: Hd => [k Hk]. exists (Z.of_nat k). by rewrite Znat.nat_N_Z -Znat.Nat2Z.inj_mul multE -Hk -pq Znat.positive_nat_Z.
      * apply/orP. move: (prime_N_divisors PRIME _ H) => [H3|H3].
        + right. rewrite -Znat.positive_nat_N in H3. by rewrite -pq (Nat2N.inj _ _ H3).
        + left. by rewrite (Nat2N.inj _ (N.of_nat 1) H3).
  Qed.

  Lemma Iq_fieldMixin : GRing.Field.mixin_of [the unitRingType of 'I_q].
  Proof.
    by move => x nzx; rewrite qualifE /= prime_coprime ?gtnNdvd ?lt0n; auto using primeq.
  Qed.

  Definition Iq_idomainMixin := FieldIdomainMixin Iq_fieldMixin.
  Canonical Iq_idomainType := Eval hnf in IdomainType 'I_q  Iq_idomainMixin.
  Canonical Iq_fieldType :=
    GRing.Field.Pack
      {|
        GRing.Field.base := GRing.IntegralDomain.class Iq_idomainType;
        GRing.Field.mixin := Iq_fieldMixin |}.

  Definition Iq_of_zp (m : zp p) : Iq_fieldType :=
    let (n,_) := m in inZp (N.to_nat n).

  Lemma smaller1 x (Hx:(x < N.pos p)%N) : N.to_nat x < q.
  Proof.
    rewrite/q' -Znat.positive_N_nat -addn2 -[N.to_nat (N.pos p') + 2]/((N.to_nat (N.pos p') + N.to_nat 2)%coq_nat) -N2Nat.inj_add. apply/ltP. rewrite -N2Nat_inj_lt. lia.
  Qed.

  Lemma inZp0 x : @inZp q'.+1 x != 0%R → x < q → x ≠ 0.
  Proof. by move => H1 H2 H3; subst x; discriminate H1. Qed.

  Lemma N2Nat_inj_mod d : forall x,
    N.to_nat (x mod (N.pos d)) = N.to_nat x %[mod Pos.to_nat d].
  Proof.
    move => [|x]; rewrite/N.modulo //=.
      move: (N.pos_div_eucl_spec x (N.pos d)). case: (N.pos_div_eucl _ _) => [s r] /=. rewrite -[Pos.to_nat x]/(N.to_nat (N.pos x)) => ->. by rewrite N2Nat.inj_add N2Nat.inj_mul /= multE plusE modnMDl.
  Qed.

  Lemma embed_divide a b : (Z.abs_nat a) %| (Z.abs_nat b) <-> (a | b)%Z.
  Proof.
    split.
    - move/dvdnP => [k Hk]. rewrite -Z.divide_abs_l -Z.divide_abs_r. exists (Z.of_nat k). rewrite 2!Znat.Zabs2Nat.abs_nat_spec in Hk. assert (H1: (0 <= Z.abs a)%Z) by apply Z.abs_nonneg. assert (H2: (0 <= Z.abs b)%Z) by apply Z.abs_nonneg. assert (H3: (0 <= Z.of_nat k)%Z) by apply Zorder.Zle_0_nat. apply Znat.Z2Nat.inj; auto.
      * by apply Z.mul_nonneg_nonneg.
      * rewrite Hk. rewrite Znat.Z2Nat.inj_mul; auto.
        by rewrite Znat.Nat2Z.id.
    - move => [k ->]. apply/dvdnP. exists (Z.abs_nat k). by rewrite Znat.Zabs2Nat.inj_mul multE.
  Qed.

  Lemma embed_gcd a b d :
    Z.abs_nat d = gcdn (Z.abs_nat a) (Z.abs_nat b) <->
    Znumtheory.Zis_gcd a b d.
  Proof.
    split.
    - move => H. constructor.
      * rewrite -embed_divide H. by apply dvdn_gcdl.
      * rewrite -embed_divide H. by apply dvdn_gcdr.
      * move => z. by rewrite -3!embed_divide H dvdn_gcd => -> ->.
    - move => [H1 H2 H3]. symmetry. apply gcdn_def.
      * by rewrite embed_divide.
      * by rewrite embed_divide.
      * move => e. by rewrite -(Znat.Zabs2Nat.id e) 3!embed_divide; apply H3.
  Qed.

  Lemma pos_lt_mul a b c : (a < b)%positive -> (a < b * c)%positive.
  Proof.
    rewrite 2!Pos.lt_iff_add => [[r <-]]. case: (Pos.succ_pred_or c) => [->|H1].
    - rewrite Pos.mul_1_r. by exists r.
    - exists ((a * (Pos.pred c)) + r * c)%positive. by rewrite Pos.mul_add_distr_r Pos.add_assoc -{1}(Pos.mul_1_r a) -Pos.mul_add_distr_l -Pplus_one_succ_l H1.
  Qed.

  Lemma embed_mod a b :
    Z.of_nat (a %% b.+1) = ((Z.of_nat a) mod (Z.of_nat b.+1))%Z.
  Proof.
    assert (H: (Z.of_nat b.+1 > 0)%Z) by lia.
    unfold Z.modulo. move: (Zdiv.Z_div_mod (Z.of_nat a) _ H). case: (Z.div_eucl _ _) => [q1 r1] [H2 [H3 H4]]. assert (H1: (0 <= q1)%Z).
    - move: H2. case: q1 => [|q1|q1] //=. move: H3 H4. case: r1 => [|r1|r1] //=.
      * move => _ _ H1. move: (Zorder.Zle_0_nat a). rewrite H1. done.
      * move => H1 H2. rewrite Z.pos_sub_lt.
        + move => H3. move: (Zorder.Zle_0_nat a). rewrite H3. done.
        + apply pos_lt_mul. by lia.
    - rewrite -(Znat.Nat2Z.id a). rewrite H2. rewrite Znat.Z2Nat.inj_add; auto.
      * rewrite Znat.Z2Nat.inj_mul; try lia.
        rewrite Znat.Nat2Z.id. rewrite multE. rewrite plusE. rewrite -modnDm. rewrite modnMr. rewrite add0n. rewrite modn_mod. rewrite modn_small.
        + by rewrite Znat.Z2Nat.id.
        + by apply Znat.Z2Nat.inj_lt in H4; first (rewrite Znat.Nat2Z.id in H4; move/ltP: H4).
      * by apply Z.mul_nonneg_nonneg.
  Qed.

  Lemma div_eucl_0 : ∀ a, Z.div_eucl a 0%Z = (0%Z,0%Z).
  Proof. by move => [||]. Qed.

  Lemma mod_divide a b d (Hd: d ≠ 0%Z) :
    (a mod d)%Z = (b mod d)%Z <-> (d | (a - b))%Z.
  Proof.
    split.
    - unfold Z.modulo. move: (Zdiv.Z_div_mod_full a _ Hd) (Zdiv.Z_div_mod_full b _ Hd). case: (Z.div_eucl a d) => [qa ra] [-> Ha]. case: (Z.div_eucl b d) => [qb rb] [-> Hb] ->. replace (d * qa + rb - (d * qb + rb))%Z with (d * (qa - qb))%Z by lia. exists (qa - qb)%Z. by lia.
    - move => [k Hk]. assert (Hk': a = (k*d+b)%Z) by lia. subst a. by rewrite Z.add_comm Zdiv.Z_mod_plus_full.
  Qed.

  Lemma abs_nat_of_N x : Z.abs_nat (Z.of_N x) = N.to_nat x.
  Proof.
    apply Znat.Nat2Z.inj. by rewrite Znat.Nat2Z.inj_abs_nat Znat.N_nat_Z Z.abs_eq; last apply Znat.N2Z.is_nonneg.
  Qed.

  Lemma mod_pos a b : (0 <= a mod (Z.pos b))%Z.
  Proof.
    unfold Z.modulo. move: (Zdiv.Z_div_mod a _ (Zorder.Zgt_pos_0 b)). by case: (Z.div_eucl _ _) => [q r] [_ [Hr _]].
  Qed.

  Import LinearCombinationSpec modpProof.

  Import FieldSpec.

  Existing Instances zpEquiv zpPlus zpZero zpOpp zpMult zpOne zpInv.

  Lemma modp_field_spec : spec Iq_fieldType Iq_of_zp.
  Proof.
    split.
    - rewrite/Iq_of_zp => x y. split.
      * by move => ->.
      * case: x => [x Hx]. case: y => [y Hy] /= => [[H]]. assert (Hx':=Hx). assert (Hy':=Hy). move/ltb_lt: Hx' => Hx'. move/ltb_lt: Hy' => Hy'. rewrite (modn_small (smaller1 Hx')) (modn_small (smaller1 Hy')) in H. move: Hx Hy. rewrite (N2Nat.inj _ _ H) => Hx Hy. by apply zp_eq.
    - by apply: val_inj.
    - apply: val_inj => /=. apply overflow_elim => z [[Hz <-]|Hz] //=; lia.
    - move => [x _] [y _] /=. apply: val_inj. rewrite /= modnDm -plusE -N2Nat.inj_add. apply: (@overflow_elim p (x+y) (fun o => N.to_nat o = N.to_nat (x + y) %[mod q])) => z [[Hz <- //=]|Hz].
      by rewrite -Hz N2Nat.inj_add /= Pos2Nat.inj_add plusE addn2 addnC -{2}(mul1n (Pos.to_nat p').+2) (modnMDl 1 (N.to_nat z) ((Pos.to_nat p').+2)).
    - move => [[|x] Hx] /=.
      * apply: val_inj; by rewrite /= !mod0n subn0 modnn.
      * apply: val_inj. rewrite /= (N2Nat.inj_sub (Npos p) (Npos x)) /= minusE (@div.modn_small (Pos.to_nat x)).
        + by rewrite pq.
        + apply: (@smaller1 (N.pos x)). by rewrite -N.ltb_lt.
    - move => x y. simpl. rewrite mod_mul_spec. move: x => [x Hx]. move: y => [ y Hy ].
      apply: val_inj. by rewrite /= modnMm -pq N2Nat_inj_mod N2Nat.inj_mul multE.
    - move => [x Hx] /= H0; move/ltb_lt: Hx => Hx. apply: val_inj. rewrite /= Znat.Z_N_nat -[((inZp (N.to_nat x))^-1)%R]/(Zp_inv (@inZp q'.+1 (N.to_nat x))) /Zp_inv (prime_coprime _ primeq) gtnNdvd.
      * rewrite /= (@modn_small (N.to_nat x) q); last by apply smaller1. assert (H10: 0 < (N.to_nat x)) by (rewrite lt0n; apply/eqP; by apply inZp0; last apply smaller1). move: (egcdnP q H10). move => [a b H11 H12] /=.
        destruct bezout as [u v d D K H1]. simpl. move: H1. rewrite -embed_gcd /= abs_nat_of_N pq => H1. apply Znat.Nat2Z.inj. rewrite 2!embed_mod -pq Znat.positive_nat_Z Znat.Z2Nat.id; last by apply mod_pos. rewrite Zdiv.Zmod_mod mod_divide // -embed_divide /= pq -(@Gauss_dvdl _ _ (N.to_nat x)).
        + apply/dvdnP. exists ((Z.abs_nat (- (v + Z.of_nat b)))). rewrite -abs_nat_of_N -{1}multE -Znat.Zabs2Nat.inj_mul Z.mul_sub_distr_r. replace (u * Z.of_N x)%Z with (d - v * Z.pos p)%Z by lia. rewrite -Znat.N_nat_Z -Znat.Nat2Z.inj_mul multE H11 -H1 -{1}plusE -{1}multE Znat.Nat2Z.inj_add Znat.Nat2Z.inj_mul Znat.Nat2Z.inj_abs_nat Z.abs_eq // -pq Znat.positive_nat_Z. replace (d - v * Z.pos p - (Z.of_nat b * Z.pos p + d))%Z with (- (v + Z.of_nat b) * Z.pos p)%Z by lia. by rewrite Znat.Zabs2Nat.inj_mul /= multE.
        + rewrite (prime_coprime _ primeq) gtnNdvd //. by apply smaller1.
      * by rewrite lt0n.
      * by apply widen_ord_proof.
  Qed.

End Zmodp.
