Require Import Utf8 ZArith.
Require modp LargePrime.
Require Import LinearCombination LinearCombinationSpec QAPproof QapGenProof modpProof project RField.
Import Monads QAP.
(* Chose LCList or LCMap to implement linear combinations. *)
Module LC <: LinearCombinationType <: LCSpec := LCListSpec.

From mathcomp
Require Import ssreflect ssrnat ssrbool eqtype fintype ssrfun seq.
From mathcomp
Require Import div bigop prime ssralg finalg zmodp poly polydiv.
Import GRing.Theory.
Require Import ssrnumbers lagrange.

Set Implicit Arguments.


(* This takes too long when there are large prime numbers in the
   context... *)
Lemma pos_succ ω : (ω <= Pos.succ ω)%positive.
Proof. Psatz.lia. Qed.


(* This section would become a functor if QAP does *)
Section QAPpoly.

  Variable p' : positive.
  Local Notation p := (p' + 2)%positive.

  Let q' := Pos.to_nat p'.
  Local Notation q := q'.+2.

  Lemma pq : Pos.to_nat p = q.
  Proof. by rewrite -addn2 -plusE -(Pos2Nat.inj_add _ 2%positive). Qed.

  Hypothesis primep : Znumtheory.prime (Zpos p).
  Let primeq := primeq primep.

  Canonical Iq_fieldType := Iq_fieldType primep.

  Definition Iq_of_nat n : Iq_fieldType := @inZp q'.+1 n.
  Definition Iq_of_pos p : 'I_q := Iq_of_nat (Pos.to_nat p).
  Local Notation Iq_of_zp := (Iq_of_zp primep).

  Lemma nat_of_ord_Iq_of_nat n (Hn:n < q) :
    nat_of_ord (Iq_of_nat n) = n.
  Proof. rewrite /Iq_of_nat /inZp /=. by apply modn_small. Qed.


  Section PolyC.

    (* The number of roots *)
    Variable next_root : positive.
    Hypothesis next_root_smaller: Pos.to_nat next_root <= q.

    Lemma root_smaller r (Hr:r < Pos.to_nat next_root) : r < q.
    Proof. by apply (@leq_trans (Pos.to_nat next_root)). Qed.

    (* The roots are between [1] and [next_root-1] (included) *)
    Definition roots_nat : seq nat :=
      index_iota 1 (Pos.to_nat next_root).
    Definition roots : seq Iq_fieldType :=
      [seq (Iq_of_nat r) | r <- roots_nat].

    (* They are distinct *)
    Lemma roots_uniq : uniq roots.
    Proof.
      rewrite map_inj_in_uniq.
      - by apply iota_uniq.
      - move => x y /=. rewrite !mem_index_iota => /andP [_ Hx] /andP [_ Hy] [Hxy]. rewrite !modn_small in Hxy; auto using root_smaller.
    Qed.

    (* The divisor polynomial *)
    Definition polyD : {poly Iq_fieldType} :=
      (\prod_(r <- roots) ('X - r%:P))%R.

    Lemma polyD_root r (Hr:r \in roots) : root polyD r.
    Proof. by rewrite root_prod_XsubC. Qed.

    (* A polynomial of a concrete QAP for which I am going to generate the
     corresponding mathematical polynomial *)
    Variable c : M.t (LC.t (modp.zp p)).


    Section PolyK.

      (* The index of the polynomial we generate *)
      Variable k : positive.

      Definition value r : Iq_fieldType :=
        match M.get (Pos.of_nat r) c with
          | Some c' => Iq_of_zp (LC.coeff c' k)
          | None => 0%R
        end.

      (* We have [next_root-1] points:
         - x is a root
         - y is the kth coeff of the linear combination at root x
      *)
      Definition points : seq (Iq_fieldType * Iq_fieldType)%type :=
        [seq (Iq_of_nat r, value r) | r <- roots_nat].

      Lemma points_roots : [seq r.1 | r <- points] = roots.
      Proof.
        assert (H:∀ m, [seq r.1 | r <- [seq (Iq_of_nat r, value r) | r <- iota m (Pos.to_nat next_root - 1)]] = [seq Iq_of_nat r | r <- iota m (Pos.to_nat next_root - 1)]); last by apply H.
        elim: (Pos.to_nat next_root - 1) => [|n IHn] //= m.
        by rewrite IHn.
      Qed.

      Lemma points_uniq : uniq [seq r.1 | r <- points].
      Proof. by rewrite points_roots; apply roots_uniq. Qed.

      Definition polyK : {poly 'I_q} := lagrange Iq_fieldType points.

      Lemma polyK_interpolation r : r \in roots →
        (polyK.[r])%R = value r.
      Proof.
        rewrite -points_roots => /mapP [[x y] Hx] /= ->.
        move: (@lagrange_interpolation Iq_fieldType _ points_uniq _ Hx) => /= ->.
        move/mapP: Hx => [a Ha] [-> ->]. rewrite nat_of_ord_Iq_of_nat //=.
        rewrite mem_index_iota in Ha. move/andP: Ha => [_ Ha]. by apply root_smaller.
      Qed.

    End PolyK.


    (* The number of wires *)
    Variable ω : positive.

    (* An assignment of wires *)
    Variable ρ : positive -> 'I_q.

    (* The associated polynomial *)
    Definition polyC : {poly 'I_q} :=
      (\sum_(1 <= k < (Pos.to_nat ω).+1) ((ρ (Pos.of_nat k)) *: (polyK (Pos.of_nat k))))%R.

    (* Its values on the roots *)
    Lemma polyC_interpolation r (Hr:r \in roots) :
      (polyC.[r])%R = (\sum_(1 <= k < (Pos.to_nat ω).+1) ((ρ (Pos.of_nat k)) * (value (Pos.of_nat k) r)))%R.
    Proof.
      rewrite horner_sum. apply eq_big; auto.
      move => k _. by rewrite hornerZ polyK_interpolation.
    Qed.

  End PolyC.


  Section PolyQ.

    (* The QAP for which we build the polynomial *)
    Variable qap : cqap (modp.zp p).
    Hypothesis next_root_smaller : Pos.to_nat (next_root qap) <= q.
    Notation ω := (next_wire qap).

    (* An assignment of its wires *)
    Variable ρ : positive -> 'I_q.

    Definition polyQ : {poly 'I_q} :=
      (((polyC (next_root qap) (cV qap) ω ρ) *
        (polyC (next_root qap) (cW qap) ω ρ)) -
       (polyC (next_root qap) (cY qap) ω ρ))%R.

    (* Its values on the roots *)
    Lemma polyQ_interpolation r (Hr:r \in roots (next_root qap)) :
      (polyQ.[r])%R =
      (((\sum_(1 <= k < (Pos.to_nat ω).+1) ((ρ (Pos.of_nat k)) * (value (cV qap) (Pos.of_nat k) r))) *
        (\sum_(1 <= k < (Pos.to_nat ω).+1) ((ρ (Pos.of_nat k)) * (value (cW qap) (Pos.of_nat k) r)))) -
       (\sum_(1 <= k < (Pos.to_nat ω).+1) ((ρ (Pos.of_nat k)) * (value (cY qap) (Pos.of_nat k) r))))%R.
    Proof. by rewrite hornerD hornerN hornerM !polyC_interpolation. Qed.

    (* What the division tells us *)
    Local Notation polyd := (polyD (next_root qap)).
    Hypothesis Hd : @dvdp Iq_fieldType polyd polyQ.

    Lemma mgates_poly r (Hr:r \in roots (next_root qap)) :
      ((\sum_(1 <= k < (Pos.to_nat ω).+1) ((ρ (Pos.of_nat k)) * (value (cV qap) (Pos.of_nat k) r))) *
       (\sum_(1 <= k < (Pos.to_nat ω).+1) ((ρ (Pos.of_nat k)) * (value (cW qap) (Pos.of_nat k) r))))%R =
      (\sum_(1 <= k < (Pos.to_nat ω).+1) ((ρ (Pos.of_nat k)) * (value (cY qap) (Pos.of_nat k) r)))%R.
    Proof.
      apply/eqP. rewrite -subr_eq0 -(polyQ_interpolation r Hr) -(andTb ((polyQ.[r])%R == 0%R)) -(polyD_root (next_root qap) r Hr) -root_gcd (eqp_root (q:=polyd)).
      - by apply polyD_root.
      - rewrite dvdp_gcd_idl.
        * reflexivity.
        * exact Hd.
    Qed.

  End PolyQ.

End QAPpoly.


Section BuildSolution.

  Local Notation p := LargePrime.large_prime.
  Local Notation p' := (p - 2)%positive.

  Lemma primep : Znumtheory.prime (Zpos (p' + 2)).
  Proof. exact large_prime_is_prime. Qed.

  Existing Instances zpEquiv zpPlus zpZero zpMult zpOne.

  (* The QAP generated by Pinocchioq *)
  Variable qap : cqap (modp.zp (p' + 2)).
  Local Notation ω := (next_wire qap).
  Hypothesis next_root_smaller : (next_root qap <= p)%positive.

  (* The valuation β comes from the crypto checks LINEAR SPAN and CONSISTENCY *)
  Variable β : positive → modp.zp p.

  (* The following hypothesis can be checked when receiving β:
       the public IOs must be 32bits integers *)
  Hypothesis IO :
    ∀ w r,
      M.get w (meta_data qap) = Some (Bank_Shared r) →
      project (β w) r ≠ None.

  (* The corresponding polynomials *)
  Definition polyq := polyQ _ primep qap (fun p => Iq_of_zp primep (β p)).
  Definition polyd := polyD _ primep (next_root qap).

  (* Hd comes from the crypto check DIVIDE *)
  Hypothesis Hd : @dvdp (Iq_fieldType _ primep) polyd polyq.

  (* Correctness of multiplication gates *)
  Lemma mgates:
    ∀ r,
      (r < (next_root qap))%positive →
      ∀ v w y,
      M.get r (cV qap) = Some v →
      M.get r (cW qap) = Some w →
      M.get r (cY qap) = Some y →
      Equivalence.equiv (mult (LC.den ω β v) (LC.den ω β w)) (LC.den ω β y).
  Proof.
    move => r Hr v w y Hv Hw Hy.
    move: (modp_field_spec primep) => H; destruct H.
    assert (H2:=@LC.denote_ad (modp.zp p) (zpPlus p) (zpZero p) (zpMult p) (Iq_fieldType _ primep) (Iq_of_zp primep) ω β).
    rewrite embed_eq embed_mult !H2.
    assert (H3: (Pos.to_nat (next_root qap) <= (Pos.to_nat p').+2)).
      - apply/leP. do 2 rewrite <- Pos2Nat.inj_succ. rewrite -Pos2Nat.inj_le. exact next_root_smaller.
    move: (mgates_poly qap H3 (fun v => Iq_of_zp primep (β v)) Hd (Iq_of_nat _ primep (Pos.to_nat r))).
    unfold value.
    assert (H4: Pos.of_nat (Iq_of_nat p' primep (Pos.to_nat r)) = r).
      - rewrite nat_of_ord_Iq_of_nat.
        * by rewrite Pos2Nat.id.
        * apply/ltP. do 2 rewrite <- Pos2Nat.inj_succ. rewrite -Pos2Nat.inj_lt. by apply (Pos.lt_le_trans _ (next_root qap)).
    rewrite H4 Hv Hw Hy => H5. apply H5. apply/mapP. exists (Pos.to_nat r); last done.
    rewrite mem_index_iota. apply/andP. split.
    - apply/ltP. by apply Pos2Nat.is_pos.
    - apply/ltP. by rewrite -Pos2Nat.inj_lt.
  Qed.

  Import Psatz.

  (* So it is a solution *)
  Lemma β_solution : solution _ (zpPlus p) (zpZero p) (zpMult p) project ω qap β.
  Proof.
    repeat split.
    - exact IO.
    - exact mgates.
    - apply pos_succ.
  Qed.

  (* Build the solution... *)
  Definition ρ (v:positive) : Z :=
    match M.get v (env qap) with
      | Some v =>
        let '(lc, r) := (lc_of_value (f_of_z:=modp.zconst _)) v in
        match project (LC.den ω β lc) r with
          | Some z => z
          | None => 0%Z
        end
      | None => 0%Z
    end.

  (* ... and prove it is a solution! *)
  Theorem ρ_solution : (cqap_solution (f_of_z:=modp.zconst _)) project qap ρ.
  Proof.
    exists β. exists ω. split.
    - exact β_solution.
    - rewrite/ρ => v lc_r /get_env_Some H z Hz. destruct lc_r as (lc & r). destruct H as (? & H & H'). by rewrite H H' Hz.
      
  Qed.

End BuildSolution.


Unset Implicit Arguments.
