Require Import Utf8.

From mathcomp
Require Import ssreflect ssrnat ssrbool eqtype fintype ssrfun seq.
From mathcomp
Require Import bigop ssralg poly.
Import GRing.Theory.

Set Implicit Arguments.


Section Util.

  Variable A : eqType.
  Variable l : seq A.
  Variables x y : A.
  Hypothesis Hx : x \in l.
  Hypothesis Hy : y \notin l.

  Lemma neqxy : x != y.
  Proof. by apply/eqP => H; subst y; move/negP: Hy. Qed.

End Util.


Arguments neqxy {A l x y} Hx Hy.


Section Lagrange.

  Variable R : fieldType.


  Section Monomial.

    Variables r p : R.

    Definition lagrange_monomial : {poly R} :=
      ((1 / (r - p)) *: ('X - p%:P))%R.

    Lemma lagrange_monomial_p :
      (lagrange_monomial.[p])%R = 0%R.
    Proof. by rewrite hornerZ hornerXsubC addrN mulr0. Qed.

    Hypothesis H : r != p.

    Lemma lagrange_monomial_size :
      size lagrange_monomial = 2.
    Proof.
      rewrite size_scale.
      - by rewrite size_XsubC.
      - by rewrite div1r invr_neq0 // subr_eq0.
    Qed.

    Lemma lagrange_monomial_nonzero :
      lagrange_monomial != 0%R.
    Proof. by rewrite -size_poly_gt0 lagrange_monomial_size. Qed.

    Lemma lagrange_monomial_r :
      (lagrange_monomial.[r])%R = 1%R.
    Proof. by rewrite hornerZ hornerXsubC div1r mulVf //= subr_eq0. Qed.

  End Monomial.


  Section Poly.

    Definition lagrange_poly (r:R) (points:seq R) :
      {poly R} :=
      (\prod_(p <- points | p != r) (lagrange_monomial r p))%R.

    Lemma lagrange_poly_nonzero r points :
      lagrange_poly r points != 0%R.
    Proof.
      rewrite/lagrange_poly. apply big_rec.
      - by apply oner_neq0.
      - move => i p Hi Hp. rewrite -size_poly_gt0. rewrite size_mul //=.
        * by rewrite lagrange_monomial_size //= eq_sym.
        * apply lagrange_monomial_nonzero. by rewrite eq_sym.
    Qed.

    Lemma lagrange_poly_notin_size r points :
      r \notin points →
      size (lagrange_poly r points) = (size points).+1.
    Proof.
      rewrite/lagrange_poly. elim: points => [|x xs IHxs] //=.
      - move => _. by rewrite big_nil size_polyC oner_neq0.
      - rewrite in_cons negb_or. move/andP => [H1 H2]. rewrite big_cons eq_sym H1 size_mul.
        * by rewrite IHxs //= lagrange_monomial_size.
        * by apply lagrange_monomial_nonzero.
        * by apply lagrange_poly_nonzero.
    Qed.

    Lemma lagrange_poly_size r points :
      uniq points → r \in points →
      size (lagrange_poly r points) = size points.
    Proof.
      rewrite/lagrange_poly. elim: points => [|s p IHp] //=.
      move/andP => [H1 H2]. rewrite in_cons big_cons. move/orP => [H|H].
      - move/eqP: H => ->. by rewrite eqxx /= lagrange_poly_notin_size.
      - rewrite eq_sym (neqxy H H1) size_mul.
        * rewrite IHp //= lagrange_monomial_size.
          + by rewrite add2n.
          + by apply (neqxy H H1).
        * apply lagrange_monomial_nonzero. by apply (neqxy H H1).
        * apply lagrange_poly_nonzero.
    Qed.

    Lemma lagrange_poly_r r points :
      ((lagrange_poly r points).[r])%R = 1%R.
    Proof.
      rewrite/lagrange_poly. apply big_rec.
      - by rewrite hornerC.
      - move => p P Hp HP. rewrite hornerM_comm.
        * rewrite lagrange_monomial_r.
          + by rewrite HP mul1r.
          + by rewrite eq_sym.
        * by apply mulrC.
    Qed.

    Lemma lagrange_poly_p r points p
          (Hp:p \in points) (H2: p != r) :
      ((lagrange_poly r points).[p])%R = 0%R.
    Proof.
      apply/eqP. rewrite horner_prod prodf_seq_eq0. apply/hasP. exists p; auto.
      apply/andP; split; auto.
      by rewrite lagrange_monomial_p.
    Qed.

  End Poly.


  Section Interpolation.

    Variable points : seq (R * R).

    Definition lagrange : {poly R} :=
      (\sum_(p <- points) (p.2 *: (lagrange_poly p.1 [seq q.1 | q <- points])))%R.

    Hypothesis H1 : uniq [seq q.1 | q <- points].

    Theorem lagrange_interpolation r (H:r \in points) :
      (lagrange.[r.1])%R = r.2.
    Proof.
      rewrite horner_sum (big_rem _ H) /= hornerZ lagrange_poly_r mulr1.
      assert (H2:(\sum_(p <- rem r points) ((p.2 *: lagrange_poly p.1 [seq q.1 | q <- points]).[r.1]))%R = 0%R).
      - rewrite big_seq. apply big_rec => //=.
        move => [x y] p Hxy -> /=. rewrite hornerZ. rewrite lagrange_poly_p.
        * by rewrite addr0 mulr0.
        * by apply map_f.
        * move: H H1 Hxy. elim: points => [|[z₁ z₂] zs IHzs] //=.
          rewrite in_cons. move/orP => [/eqP Hr|Hr].
          + subst r. rewrite eqxx. move/andP => [H10 H2] H3 /=. apply (map_f (@fst R R)) in H3. simpl in H3. rewrite eq_sym. apply (neqxy H3 H10).
          + move/andP => [H10 H2]. replace ((z₁, z₂) == r) with false.
            - rewrite in_cons. move/orP => [/eqP [H3 H4]|H3]; last auto.
              subst x y. apply (map_f (@fst R R)) in Hr. apply (neqxy Hr H10).
            - case er: r => [r₁ r₂]. subst r. symmetry. apply/eqP => [[H3 H4]]. subst r₁ r₂. apply (map_f (@fst R R)) in Hr. simpl in Hr. apply negbTE in H10. by rewrite Hr in H10.
      - by rewrite H2 addr0.
    Qed.

    Theorem lagrange_size : size lagrange <= size points.
    Proof.
      apply (leq_trans (size_sum _ _ _)).
      rewrite big_seq. apply big_rec.
      - by apply leq0n.
      - move => [x y] n Hxy Hn /=. rewrite geq_max. apply/andP. split; auto.
        apply (leq_trans (size_scale_leq _ _)). rewrite lagrange_poly_size //=.
        * by rewrite size_map.
        * rewrite -[x]/((x,y).1). by apply map_f.
    Qed.

    Theorem lagrange_uniq L
            (H2: ∀ r, r \in points → (L.[r.1])%R = r.2)
            (H3: size L <= size points) :
      L = lagrange.
    Proof.
      case e: ((L - lagrange)%R == 0%R).
      - rewrite subr_eq0 in e. by move/eqP: e.
      - assert (H: all (root (L - lagrange)) [seq q.1 | q <- points]).
        * apply/allP => x Hx. apply/rootP. rewrite hornerD hornerN. move/mapP: Hx => [y Hy ->]. by rewrite (H2 _ Hy) (lagrange_interpolation _ Hy) addrN.
        * apply negbT in e. assert (H4:=max_poly_roots e H H1). rewrite size_map in H4. assert (H5: size (L - lagrange)%R <= size points).
          + apply (leq_trans (size_add _ _)). rewrite geq_max. apply/andP. split; auto.
            by rewrite size_opp; apply lagrange_size.
          + assert (H6:=leq_ltn_trans H5 H4). by rewrite ltnn in H6.
    Qed.

  End Interpolation.

End Lagrange.


Unset Implicit Arguments.
